/* $Id: Helper.java,v 1.1 2004/03/22 19:31:49 hnetynka Exp $ */
package SOFA.Connector.Boot;

import java.lang.reflect.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

import SOFA.Connector.ConnectorTransportException;

/** Class with helper static methods. 
  * 
  * @author Petr Hnetynka
  */
public class Helper {

  private static final String GENERCONNPACKAGE = "SOFA.GeneratedConnectors";

  /** Refernce to CCNS 
    */
  public static SOFA.SOFAnode.Run.CCNS.ConnectorCacheNS ccns;

  static {
    try {
      ccns=SOFA.Connector.CCNSAccess.CCNSAccess.openCCNS();
    } catch (SOFA.Connector.CCNSAccess.CCNSAccessException e) {
      throw new RuntimeException("Cannot access CCNS.", e);
    }
  }

  /** Creates client side part of the connector for specified CDL type and the reference.
    * 
    * @param cdlName cdl full name
    * @param ref SOFA reference to the connector (server side)
    * @return client side of the connector
    */
  public static SOFA.Connector.RoleBase createCltConn(String cdlName, SOFA.Connector.Reference ref) {
    try {
      SOFA.Connector.ECG.DeploymentDescriptor connDD = createCltConnDeplDesc(cdlName, ref);
      String type = connDD.getConnectorBuilder();
      Class cl = Class.forName(type, true, SOFA.SOFAnode.Run.SOFACompClassLoader.getCompClassLoader());
      Class[] par = new Class[2];
      par[0] = Class.forName("SOFA.Connector.ECG.DeploymentDescriptor");
      par[1] = Class.forName("java.lang.ClassLoader");
      Method m = cl.getMethod("createConnector", par);
      Object[] arg = new Object[2];
      arg[0] = connDD;
      arg[1] = SOFA.SOFAnode.Run.SOFACompClassLoader.getCompClassLoader();
      SOFA.Connector.RoleBase conn = (SOFA.Connector.RoleBase) m.invoke(null, arg);
//      conn.setThrIDRegistry(thrRegistry);
//      conn.setRegister(this);
      conn.link(ref);
      return conn;
    } catch (ClassNotFoundException e)  {
      throw new ConnectorTransportException("ClassNotFoundException: "+e.getMessage(), e);
    } catch (NoSuchMethodException e)  {
      throw new ConnectorTransportException("NoSuchMethodException: "+e.getMessage(), e);
    } catch (IllegalAccessException e)  {
      throw new ConnectorTransportException("IllegalAccessException: "+e.getMessage(), e);
    } catch (java.lang.reflect.InvocationTargetException e)  {
      throw new ConnectorTransportException("InvocationTargetException: "+e.getMessage(), e);
    } catch (SOFA.Connector.LinkException e)  {
      throw new ConnectorTransportException("SOFA.Connector.LinkException: "+e.getMessage(), e);
    }
  }

  /** Returns full name of the CDL object from which the object was generated.
    * 
    * @param obj object
    * @return CDL full name or null;    
    */
  public static String getCDLNameOfObject(Object obj) {
    try {
      return (String) (obj.getClass().getField("_CDLID").get(obj));
    } catch (Exception e) {
      return null;
    }
  }

  /** Creates connector deployment descriptor for client side of the connector
    * to the specified cdl type.
    * 
    * @param cdlName cdl full name
    * @param ref SOFA reference to the connector (server side)
    * @return connector depl. descriptor
    */
  public static SOFA.Connector.ECG.DeploymentDescriptor createCltConnDeplDesc(String cdlName, SOFA.Connector.Reference ref) {
    try {
      Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
      Element root = doc.createElement("sofa_connector");
      doc.appendChild(root);
      Element el;
      String hash;
      if (ref.conn_type.equals("SOFA.Connector.ECG.SGenerator.Connectors.CSProcCallImpl")) {
        root.setAttribute("name", "NAME");
        root.setAttribute("unit", "Client");
        root.setAttribute("impl", "SOFA.Connector.ECG.SGenerator.Connectors.CSProcCallImpl");
        el = doc.createElement("connector_element");
        el.setAttribute("name", "cRole");
        if ((hash = translateMangled("CRoleImpl?"+cdlName))==null)
          throw new ConnectorTransportException("No cRole for reference "+ref);
        el.setAttribute("impl", GENERCONNPACKAGE + "."+hash+".CRoleImpl");
        root.appendChild(el);
        for (int i=0; i<ref.profiles.length; i++) {
          if (ref.profiles[i] instanceof SOFA.Connector.ECG.Profiles.LocalProfile) {
            el = doc.createElement("connector_element");
            el.setAttribute("name", "stub");
            if ((hash = translateMangled("LocalStub?"+cdlName))==null)
              throw new ConnectorTransportException("No localStub for reference "+ref);
            el.setAttribute("impl", GENERCONNPACKAGE + "."+hash+".LocalStub");
            root.appendChild(el);
          } else if (ref.profiles[i] instanceof SOFA.Connector.ECG.Profiles.RMIProfile) {
            el = doc.createElement("connector_element");
            el.setAttribute("name", "stub");
            if ((hash = translateMangled("RMIStub?"+cdlName))==null)
              throw new ConnectorTransportException("No RMIStub for reference "+ref);
            el.setAttribute("impl", GENERCONNPACKAGE + "."+hash+".RMIStub");
            root.appendChild(el);
          } else {  // CORBA stub
            el = doc.createElement("connector_element");
            el.setAttribute("name", "stub");
            if ((hash = translateMangled("CORBAStub?"+cdlName))==null)
              throw new ConnectorTransportException("No CORBAStub for reference "+ref);
            el.setAttribute("impl", GENERCONNPACKAGE + "."+hash+".CORBAStub");
            root.appendChild(el);
          }
        }
      } else if (ref.conn_type.equals("SOFA.Connector.ECG.SGenerator.Connectors.EventPassingImpl")) {
        throw new ConnectorTransportException("No suport for dynamic addition to EventPassing connector");
      } else if (ref.conn_type.equals("SOFA.Connector.ECG.SGenerator.Connectors.DataStreamImpl")) {
        throw new ConnectorTransportException("No suport for dynamic addition to DataStream connector");
      } else {
        throw new ConnectorTransportException("Unknown connector: "+ref.conn_type);
      }
      return new SOFA.Connector.ECG.DeploymentDescriptor(root);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      throw new ConnectorTransportException("Cannot create connector descriptor", e);
    } catch (java.rmi.RemoteException e) {
      throw new ConnectorTransportException("RemoteException occured", e);
    } catch (SOFA.Connector.DeploymentDescriptorException e) {
      throw new ConnectorTransportException("Cannot create connector descriptor", e);
    }
  }

  /** Translated mangled version of the name.
    * 
    * @param mangled name to be translated
    * @return translated name
    * @throws java.rmi.RemoteException remote exception
    */
  public static String translateMangled(String mangled) throws java.rmi.RemoteException {
    StringBuffer buf=new StringBuffer();
    int i, len=mangled.length();
    char a;

    buf.append("EEM__");

    for (i=0;i<len;i++) {
      a=mangled.charAt(i);
      if (a=='_') {
        buf.append("__");
      } else if (a==':') {
        if (i+1<len && mangled.charAt(i+1)==':') {
          i++;
          buf.append("_CC_");
        } else {
          buf.append("_C_");
        }
      } else if (a=='/') {
        buf.append("_S_");
      } else if (a=='.') {
        buf.append("_D_");
      } else if (a=='?') {
        buf.append("_Q_");
      } else if (a=='!') {
        buf.append("_E_");
      } else if (Character.isJavaIdentifierPart(a)) {
        buf.append(a);
      } else {
        buf.append("_"+Integer.toString(a,16)+"_");
      }
    }
    return ccns.getHashByName2(buf.toString());
  }
}
