/*
 * DataStreamLocalSenderGen.java
 *
 * Created on 13. duben 2002, 21:08
 */

package SOFA.Connector.EEG.EEM.Elements;

import SOFA.SOFAnode.Made.TIR.*;
import SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods;
import SOFA.Connector.EEG.CodeWriter.*;
import SOFA.Connector.EEG.EEM.Actions.*;

/**
 *
 * @author  ghort
 * @version 
 */
public class DataStreamLocalSenderGen extends JImplWriterDirectorBase {

	protected String is, os, bh, isPackage, osPackage;

	protected static final int pullMode=0;
	protected static final int pushMode=1;
	
	protected int sMode, rMode;

	/** Creates new DataStreamLocalSenderGen */
	public DataStreamLocalSenderGen(String provIfaceCDL, SOFA.Connector.Property[] params) throws JWriterDirectorException {
		super(provIfaceCDL,params);
		isPackage=SOFA.Connector.Property.findFirst(params,"isPackage");
		osPackage=SOFA.Connector.Property.findFirst(params,"osPackage");
		is=isPackage+".CNInterface";
		os=osPackage+".CNInterface";
		bh="SOFA.Connector.EEG.Types.DataStream.bytesHolder";

		String mode=SOFA.Connector.Property.findFirst(params,"senderMode");
		if (mode!=null) {
			if (mode.equals("push")) {
				sMode=pushMode;
			} else if (mode.equals("pull")) {
				sMode=pullMode;
			} else {
				throw new JWriterDirectorException("Parameter senderMode is incorrect '"+mode+"'.");
			}
		} else {
			throw new JWriterDirectorException("Parameter senderMode is not specified");
		}

		mode=SOFA.Connector.Property.findFirst(params,"receiverMode");
		if (mode!=null) {
			if (mode.equals("push")) {
				rMode=pushMode;
			} else if (mode.equals("pull")) {
				rMode=pullMode;
			} else {
				throw new JWriterDirectorException("Parameter receiverMode is incorrect '"+mode+"'.");
			}
		} else {
			throw new JWriterDirectorException("Parameter receiverMode is not specified");
		}

		if (sMode==pullMode && rMode==pushMode) {
			throw new JWriterDirectorException("Mode sender-push/receiver-pull is not supported.");
		}
	}

	public String[] getImplements() {
		if (sMode==pullMode) {
			return new String[] {"SOFA.Connector.EEG.Types.Sender",is,"SOFA.Connector.Linkable"};
		} else {
			return new String[] {"SOFA.Connector.EEG.Types.Sender",os};
		}
	}
	
	public String getClassBody() throws JWriterDirectorException {
		String methods;
		if (sMode==pushMode) {
			methods=
				"\tpublic void write(byte[] buffer, int len, String thr_id) {\n"+
        "\t\treceiverWriteParams[0]=buffer;\n"+
        "\t\treceiverWriteParams[1]=new Integer(len);\n"+
        "\t\treceiverWriteParams[2]=thr_id;\n"+
        "\t\ttry {\n"+
				"\t\t\treceiverWrite.invoke(receiver,receiverWriteParams);\n"+
        "\t\t} catch (IllegalAccessException _e) {\n"+
        "\t\t\tthrow new SOFA.Connector.ConnectorTransportException(\"IllegalAccessException\", _e);\n"+
        "\t\t} catch (java.lang.reflect.InvocationTargetException _e) {\n"+
        "\t\t\tthrow new SOFA.Connector.ConnectorTransportException(\"InvocationTargetException\", _e);\n"+
        "\t\t}\n"+
				"\t}\n\n";
		} else if (sMode==pullMode && rMode==pullMode) {
			methods=
				"\tpublic int read("+bh+" bufferHolder, int len, String thr_id) {"+
				"\t\treturn client.read(bufferHolder,len,thr_id);\n"+
				"\t}\n\n";
		} else {
			throw new JWriterDirectorException("Mode sender-push/receiver-pull is not supported.");
		}

		return managerTieImpl() + "\n" +
      (sMode==pushMode ? "\tprotected java.lang.reflect.Method receiverWrite;\n"+
			"\tprotected Object receiverWriteParams[]=new Object[3];\n" : "" ) +
			"\tprotected Object receiver;\n"+
			"\tprotected "+is+" client;\n\n"+
			"\tpublic void link(Object target) throws SOFA.Connector.LinkException {\n"+
			"\t\tif (target instanceof SOFA.Connector.TaggedProfile) {\n"+
			"\t\t\treceiver=((SOFA.Connector.ECG.Profiles.LocalProfile)target).link;\n"+
      (sMode==pushMode ? "\t\t\ttry {\n"+
      "\t\t\t\treceiverWrite=receiver.getClass().getMethod(\"write\",new Class[] {byte[].class, Integer.TYPE, String.class});\n"+
      "\t\t\t} catch (NoSuchMethodException e) {\n"+
      "\t\t\t\tthrow new SOFA.Connector.LinkException(\"Cannot create dynamic local binding\",e);\n"+
      "\t\t\t}\n" : "" ) +
			(sMode==pullMode && rMode==pullMode ? "\t\t\t((SOFA.Connector.EEG.EEM.Elements.DataStreamLocalNotifiable)receiver).linkNotification(this);\n" : "")+
			"\t\t} else {\n"+
			"\t\t\tclient=("+is+")target;\n"+
			"\t\t}\n"+
			"\t}\n\n"+
			"\tpublic String[] getSupportedProfiles() {\n"+
			"\t\treturn new String[] {\"LOCAL\"};\n"+
			"\t}\n\n"+methods+"\n\n"+
                        "\tpublic org.objectweb.fractal.api.Type getFcItfType() { return null; }\n"+
                        "\tpublic boolean isFcInternalItf() { return false; }\n"+
                        "\tpublic String getFcItfName() { throw new SOFA.Connector.ConnectorTransportException(\"Interconection via EventPassing connector - getFcItfName is not supported.\"); }\n"+
                        "\tpublic org.objectweb.fractal.api.Component getFcItfOwner() { throw new SOFA.Connector.ConnectorTransportException(\"Interconection via EventPassing connector - getFcItfOwner is not supported.\"); }\n\n";
	}
	
	public String getMethodDef(OperationDef mt) throws JWriterDirectorException {
		return null;
	}

	public boolean defaultImplements() {
		return false;
	}
}
