/* $Id: CompArchitecture.java,v 1.1.1.1 2003/02/11 16:19:40 bures Exp $ */
package SOFA.SOFAnode.Made.CDL;
import SOFA.SOFAnode.Made.TIR.*;
import SOFA.SOFAnode.Util.ConnectorProtocolGen.*;
import java.rmi.RemoteException;

class CompArchitecture extends CompContainedContainerIm {

  boolean isSystem;

  EnumList propNames;
  CompReffer frame;

  int numOfBind;

  java.util.ArrayList provReq;
  String[] provOpArray;  // filled by createOpArrays (depends on provReq)
  String[] allOpArray;   // filled by createOpArrays (depends on provReq)

  public CompArchitecture(ID id, CompContainer in, CompRepository inRep, CompReffer frame, EnumList propNames, boolean sys) {
    super(ObjectsKind.o_Architecture, id, in, inRep);
    this.frame = frame;
    this.propNames = propNames;
    numOfBind = 0;
    isSystem = sys;
//    Debug.message("Vytvarim architecturu: "+id);
  }

  public void checkConsist(EnumList props, CompRepository rep) throws CDLExceptCheck, CDLExceptLock, CDLExceptRemote {
//    Debug.message("testuji konzistenci architectury: "+getIdent());
    CompContained[] cont = contents();
    for (int i=0;i<cont.length;i++) {
      try {
        cont[i].checkConsist(propNames, rep);
        //
        //if (cont[i].objectKind() == ObjectsKind.o_Inst) 
        //  Debug.message(((CompInst)cont[i]).provOps.toString());
      } catch (CDLExceptCheck e) {
        throw new CDLExceptCheck(e.getMessage()+" in architecture "+fullName());
      }
    }
    DetectedConnectors dc; 
    try {
      dc = detectConnectors();
    } catch (BadBindException e) {
      Output.out.println(e.getMessage());
      Output.out.println(e.getCause().getMessage());
      return;
    }
    if (rep.testProtocols && (this.cont.size()!=0)) {
      Output.out.println("Checking protocol of architecture: "+getIdent());

      String archP = generateArchProtocol(dc);
      if (archP==null || archP.compareTo("")==0) {
        Output.out.println("Warning: frame of architecture "+getIdent()+" is empty, protocols can't be checked.");
        return;
      }
      String instP = "";
      try {
        instP = generateInstProtocol(dc);
      } catch (BadBindException e) {
        Output.out.println("Warning: in architecture "+getIdent()+", there is bad binds. Protocols can't be checked.\n * "+e.getMessage());
        return;
      } catch (CannotGenerateException e) {
        Output.out.println("Warning: protocols in architecture "+getIdent()+" can't be checked.\n * "+e.getMessage());
        return;
      }
      createOpArrays(); 
      String res;
      try {
        if (System.getProperty("cdl.debug","no").compareTo("yes")==0) {
          res = SOFA.SOFAnode.Util.ProtocolChecker.frame2arch(archP, provOpArray, instP, allOpArray);
        }
        long tmb = java.util.Calendar.getInstance().getTime().getTime();
        res = repository().protocolChecker.frame2arch(archP, provOpArray, instP, allOpArray);
        long tme = java.util.Calendar.getInstance().getTime().getTime();
        Output.out.println("Time of protocol checking of architecture: "+getIdent().name+": "+Long.toString(tme-tmb)+" ms");
        if (res != null)
          Output.out.println("Warning: "+res+" ("+fullName()+")"); 
      } catch (SOFA.SOFAnode.Util.SyntaxErrorException e) {
        Output.out.println("Protocol syntax error\n"+e.getMessage());
      }
      System.gc();
    }
  }

  public void addToNormal(Container obj, CDLRepository normRep, java.util.ArrayList delayed) throws CDLExceptToNormal, CDLExceptRemote, CDLExceptLock {
    if (id.isin == IDKind.versiontag) {
      try {
        ((Contained)obj).tag(id.tag);
      } catch (RemoteException e) {
        throw new CDLExceptRemote("Remote exception occured: "+e.getMessage());
      } catch (TIRExceptLock e) {
        throw new CDLExceptLock("Repository is locked.");
      } catch (TIRExceptCreate ecr) {
        throw new CDLExceptToNormal("Can't set tag for "+ fullName());
      }
    }
    cont.toFirst();
    try {
      for (int i=0;i<cont.size(); i++) {
        CompContained akt = (CompContained) cont.aktual();
        switch (akt.objectKind()) {
        // *****************
        case ObjectsKind.o_Inst:
          try {
            InstDef inst = ((ArchitectureDef) obj).create_inst(akt.getIdent().name, ((CompInst)akt).subcom.toNormal(normRep), ((CompInst)akt).mode);
            if (inst == null) 
              throw new CDLExceptToNormal("Can't create instance "+((CompContained)akt).fullName());
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create instance "+((CompContained)akt).fullName());
          }
          break;
        // *****************
        case ObjectsKind.o_Bind:
          try {
            BindOperDef lhs = ((CompBind) akt).lhs.toNormal((ArchitectureDef) obj, normRep);
            BindOperDef rhs = ((CompBind) akt).rhs.toNormal((ArchitectureDef) obj, normRep);
            BindDef bnd = ((ArchitectureDef) obj).create_bind(lhs,rhs,((CompBind) akt).mode, ((CompBind) akt).using);
            if (bnd==null)
              throw new CDLExceptToNormal("Can't create bind "+((CompContained)akt).fullName());
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create bind "+((CompContained)akt).fullName());
          }
          break;
        // *****************
        case ObjectsKind.o_Property:
          try {
            PropertyDef prp = ((ArchitectureDef) obj).create_property(akt.getIdent().name,((CompProperty)akt).tp.toNormal(normRep) );
            if (prp == null)
              throw new CDLExceptToNormal("Can't create property "+((CompContained)akt).fullName());
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create property "+((CompContained)akt).fullName()+"\n"+ ecr.getMessage() );
          }
          break;
        default:
          throw new CDLExceptToNormal("Unexcepted kind of object");
        }
        cont.toNext();
      }
    } catch (RemoteException e) {
      throw new CDLExceptRemote("Remote exception occured: "+e.getMessage());
    } catch (TIRExceptLock e) {
      throw new CDLExceptLock("Repository is locked.");
    }
  }

  /* from CompContainedContainerIm */
  public boolean addObject(CompContained o) {
    if (!super.addObject(o))
      return false;
    if (o instanceof CompProperty) {
      propNames.addName(((CompProperty) o).getIdent().name);
    }
    return true;
  }
  
  static class reProvReq {
    String orig;
    String rew;
    int type;
    public reProvReq(String o, int t) {
      orig = o;
      type = t;
    }
  }
  
  public String generateArchProtocol(DetectedConnectors dc) throws CDLExceptLock, CDLExceptRemote {
    int i,j;
    if (frame.isNew) {  // frame is currently compiled
      CompFrame fr = (CompFrame) frame.getRefferedObject(repository());
      provReq = new java.util.ArrayList();
      //java.util.ArrayList ln = new java.util.ArrayList();
      CompContained[] cont = fr.contents();
      for (i=0;i<cont.length;i++) {
        if (cont[i].objectKind() == ObjectsKind.o_Provides) {
          /*if (((CompProvide)cont[i]).typeCode.indexOf('a')!= -1) {
            return null;
          }*/
          provReq.add(new reProvReq(new String(cont[i].getIdent().name), ObjectsKind.o_Provides));
        }
        if (cont[i].objectKind() == ObjectsKind.o_Requires) {
          /*if (((CompRequire)cont[i]).typeCode.indexOf('a')!= -1) {
            return null;
          }*/
          provReq.add(new reProvReq(new String(cont[i].getIdent().name), ObjectsKind.o_Requires));
        }
      }
      rewriteProt(provReq, dc);
      return compFrameProtToArch(fr.protocol,provReq);
    } else {  // frame is in repository
      try {
        FrameDef fr = (FrameDef) frame.getRefferedObject(repository());
        ProtocolDef protocol = fr.protocol();
        if (protocol == null)
          return null;
        provReq = new java.util.ArrayList();
        Contained[] contRep = fr.contents(null);
        for (i=0;i<contRep.length;i++) {
          if (contRep[i].get_def_kind().value() == DefinitionKind.dk_Provides) {
            provReq.add(new reProvReq(new String(contRep[i].get_identification().name()), DefinitionKind.dk_Provides));
          }
          if (contRep[i].get_def_kind().value() == DefinitionKind.dk_Requires) {
            provReq.add(new reProvReq(new String(contRep[i].get_identification().name()), DefinitionKind.dk_Requires));
          }
        }
        rewriteProt(provReq, dc);
        return normFrameProtToArch(protocol.protocol(), provReq);
      } catch (RemoteException e) {
        throw new CDLExceptRemote(e.getMessage());
      } catch (TIRExceptLock e) {
        throw new CDLExceptLock(e.getMessage());
      }
    }
  }

  /** called from generateArchProtocol. Method assign "new" names for interfaces in
    * architecture protocol. */
  void rewriteProt(java.util.ArrayList provReq, DetectedConnectors dc) {
    int i,j;
    CompContained[] cont = contents();  // get content of this architecture
    for(i=0;i<cont.length;i++) {
      if (cont[i].objectKind() == ObjectsKind.o_Bind) { //this is bind
        CompBind bind = (CompBind) cont[i];
        String t,r1,r2,l1,l2;
        CompBindOper rhs, lhs;
        int index;
        DetectedConnector cn;
        switch (bind.mode) {
          case CompBind.m_delegate:
            lhs = bind.lhs;
            rhs = bind.rhs;
            cn = dc.containsConnWithSRole(rhs);
            index = cn.indexOfCRole(lhs);
            lhs.elements.toFirst();
            t = ((CompBindType) lhs.elements.aktual()).name;
            for(j=0;j<provReq.size();j++) {
              if ( t.compareTo(((reProvReq)(provReq.get(j))).orig) == 0 ) {
                rhs.elements.toFirst();
                rhs.subs.toFirst();
                r1 = ((CompBindType) rhs.subs.aktual()).name;
                r2 = ((CompBindType) rhs.elements.aktual()).name; 
                ((reProvReq)(provReq.get(j))).rew = "<"+t+"-"+r1+":"+r2+">";
              /*  ((reProvReq)(provReq.get(j))).rew = "<"+t+"-cRole"+Integer.toString(index)+">"; */
                break;
              }
            }
            break;
          case CompBind.m_subsume:
            rhs = bind.rhs;
            cn = dc.containsConnWithSRole(rhs);
            rhs.elements.toFirst();
            t = ((CompBindType) rhs.elements.aktual()).name;
            for(j=0;j<provReq.size();j++) {
              if ( t.compareTo(((reProvReq)(provReq.get(j))).orig) == 0 ) {
                lhs = bind.lhs;
                lhs.elements.toFirst();
                lhs.subs.toFirst();
                l1 = ((CompBindType) lhs.subs.aktual()).name;
                l2 = ((CompBindType) lhs.elements.aktual()).name;
                ((reProvReq)(provReq.get(j))).rew = "<"+l1+":"+l2+"-"+t+">";
                /*((reProvReq)(provReq.get(j))).rew = "<sRole-"+t+">"; */
                break;
              }
            }
            break;
          case CompBind.m_bind:
            if (bind.rhs.kind==CompBindOper.b_provreq && bind.lhs.kind==CompBindOper.b_provreq) {
              cn = dc.containsConnWithSRole(bind.rhs);
              index  = cn.indexOfCRole(bind.lhs);
              bind.rhs.elements.toFirst();
              String rname = ((CompBindType)bind.rhs.elements.aktual()).name;
              bind.lhs.elements.toFirst();
              String lname = ((CompBindType)bind.lhs.elements.aktual()).name;
              for(j=0;j<provReq.size();j++) {
                /*if ( rname.compareTo(((reProvReq)(provReq.get(j))).orig) == 0 || lname.compareTo(((reProvReq)(provReq.get(j))).orig) == 0) {
                  ((reProvReq)(provReq.get(j))).rew = "<"+((reProvReq)(provReq.get(j))).orig+">";
                }*/
                if ( rname.compareTo(((reProvReq)(provReq.get(j))).orig) == 0 ) {
                  ((reProvReq)(provReq.get(j))).rew = "<sRole0-"+rname+">";
                }
                if (lname.compareTo(((reProvReq)(provReq.get(j))).orig) == 0) {
                   ((reProvReq)(provReq.get(j))).rew = "<"+lname+"-cRole"+Integer.toString(index)+">";
                }
              }
            }
            break;
        } //-- end of switch
      }
    }
  }

  public static String compFrameProtToArch(CompProtOper p, java.util.ArrayList names) {
    StringBuffer sb = new StringBuffer();
    int i;
    switch (p.type) {
    case CompProtKind.p_operand:
      CompProtOperand op = (CompProtOperand) p;
      if (op.mode == CompProtMode.pm_in)
        sb.append("?");
      else
        sb.append("!");
      for (i=0;i<names.size();i++) {
        if ( op.ifacename.compareTo(((reProvReq)(names.get(i))).orig) == 0 ) {
          sb.append(((reProvReq)(names.get(i))).rew);
        }
      }
      sb.append(".");
      sb.append(((CompProtOperand) p).operand);
      break;
    case CompProtKind.p_nulloperand:
      sb.append("NULL");
      break;
    case CompProtKind.p_metaoperand:
      sb.append("[");
      sb.append(((CompProtMetaOperand) p).name);
      sb.append("]");
      break;
    case CompProtKind.p_nestedoperand:
      CompProtNestedOperand nop = (CompProtNestedOperand) p;
      if (nop.mode == CompProtMode.pm_in)
        sb.append("?");
      else
        sb.append("!");
      for (i=0;i<names.size();i++) {
        if ( nop.ifacename.compareTo(((reProvReq)(names.get(i))).orig) == 0 ) {
          sb.append(((reProvReq)(names.get(i))).rew);
          break;
        }
      }
      sb.append(".");
      sb.append(nop.operand);
      sb.append("{");
      sb.append(compFrameProtToArch(nop.nestOperand, names));
      sb.append("}");
      break;
    case CompProtKind.p_repetition:
      sb.append("(");
      sb.append(compFrameProtToArch(((CompProtUnOper) p).operand, names));
      sb.append(")*");
      break;
    case CompProtKind.p_andparallel:
      sb.append("(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand1, names));
      sb.append(")|(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand2, names));
      sb.append(")");
      break;
    case CompProtKind.p_orparallel:
      sb.append("(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand1, names));
      sb.append(")||(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand2, names));
      sb.append(")");
      break;
    case CompProtKind.p_alternative:
      sb.append("(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand1, names));
      sb.append(")+(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand2, names));
      sb.append(")");
      break;
    case CompProtKind.p_sequence:
      sb.append("(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand1, names));
      sb.append(");(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand2, names));
      sb.append(")");
      break;
    case CompProtKind.p_restriction:
      sb.append("(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand1, names));
      sb.append(")/(");
      sb.append(compFrameProtToArch(((CompProtBinOper) p).operand2, names));
      sb.append(")");
      break;
    default: break;
    }
    return sb.toString();
  }

  public static String normFrameProtToArch(ProtocolOperDef p, java.util.ArrayList names) throws RemoteException {
    StringBuffer sb = new StringBuffer();
    int i;
    String ifacename;
    switch (p.get_po_kind().value()) {
    case ProtocolOperKind.pok_operand:
      ProtocolOperandDef op = (ProtocolOperandDef) p;
      if (op.operand_mode().value() == ProtocolMode.pm_in)
        sb.append("?");
      else
        sb.append("!");
      ifacename = op.operand()[0];
      for (i=0;i<names.size();i++) {
        if ( ifacename.compareTo(((reProvReq)(names.get(i))).orig) == 0 ) {
          sb.append(((reProvReq)(names.get(i))).rew);
        }
      }
      sb.append(".");
      sb.append(op.operand()[1]);
      break;
    case ProtocolOperKind.pok_nulloperand:
      sb.append("NULL");
      break;
    case ProtocolOperKind.pok_metaoperand:
      sb.append("[");
      sb.append(((ProtocolMetaOperandDef)p).name());
      sb.append("]");
      break;
    case ProtocolOperKind.pok_nestedoperand:
      ProtocolNestedOperandDef nop = (ProtocolNestedOperandDef) p;
      if (nop.operand_mode().value() == ProtocolMode.pm_in)
        sb.append("?");
      else
        sb.append("!");
      ifacename = nop.operand()[0];
      for (i=0;i<names.size();i++) {
        if ( ifacename.compareTo(((reProvReq)(names.get(i))).orig) == 0 ) {
          sb.append(((reProvReq)(names.get(i))).rew);
        }
      }
      sb.append(".");
      sb.append(nop.operand()[1]);
      sb.append("{");
      sb.append(normFrameProtToArch(nop.nested_operand(), names));
      sb.append("}");
      break;
    case ProtocolOperKind.pok_repetition:
      sb.append("(");
      sb.append(normFrameProtToArch(((ProtocolUnOperationDef) p).operand(), names));
      sb.append(")*");
      break;
    case ProtocolOperKind.pok_andparallel:
      sb.append("(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand1(), names));
      sb.append(")|(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand2(), names));
      sb.append(")");
      break;
    case ProtocolOperKind.pok_orparallel:
      sb.append("(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand1(), names));
      sb.append(")||(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand2(), names));
      sb.append(")");
      break;
    case ProtocolOperKind.pok_alternative:
      sb.append("(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand1(), names));
      sb.append(")+(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand2(), names));
      sb.append(")");
      break;
    case ProtocolOperKind.pok_sequence:
      sb.append("(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand1(), names));
      sb.append(");(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand2(), names));
      sb.append(")");
      break;
    case ProtocolOperKind.pok_restriction:
      sb.append("(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand1(), names));
      sb.append(")/(");
      sb.append(normFrameProtToArch(((ProtocolBinOperationDef) p).operand2(), names));
      sb.append(")");
      break;
    default: break;
    }
    return sb.toString();
  }

  private static class arrayOfArraysOfInsts extends java.util.ArrayList {
    public arrayOfInsts myGet(int i) {
      return (arrayOfInsts) get(i);
    }
    public int containsName(String name) {
      for (int i=0;i<size();i++) {
        if (myGet(i).containsName(name) != -1)
          return i;
      }
      return -1;
    }
    public String protocol() {
      StringBuffer ret = new StringBuffer(myGet(0).protocol());
      for (int i=1;i<size();i++) {
        ret.append("&");
        ret.append(myGet(i).protocol());
      }
      return ret.toString();
    }
  }

  private static class arrayOfInsts extends java.util.ArrayList {
    public protInst myGet(int i) {
      return (protInst) get(i);
    }
    public int containsName(String name) {
      for (int i=0;i<size();i++) {
        if (myGet(i).inst.getIdent().name.compareTo(name)==0)
          return i;
      }
      return -1;
    }
    public String protocol() {
      StringBuffer ret = new StringBuffer(myGet(0).protocol());
      for (int i=1;i<size();i++) {
        ret.append("&");
        ret.append(myGet(i).setWithProtocol());
      }
      return ret.toString();
    }
    public void appendAI(arrayOfInsts a) {
      for(int i=0;i<a.size();i++) {
        add(a.myGet(i));
      }
    }
  }

  private static class protInst {
    CompInst inst;
    java.util.ArrayList set;   //list of Strings
    public protInst(CompInst i) {
      inst = i;
      set = new java.util.ArrayList();
    }
    public String protocol() {
      return inst.protocol;
    }
    public String setWithProtocol() {
      StringBuffer ret = new StringBuffer();
      if (set.size()==0)
        return inst.protocol;
      for (int i=0;i<set.size();i++) {
        ret.append((String)set.get(i));
        if (i!=set.size()-1)
          ret.append(",");
      }
      ret.append("&");
      ret.append(inst.protocol);
      return ret.toString();
    }
    public void addToSet(String[] a) {
      for(int i=0;i<a.length;i++) {
        set.add(a[i]);
      }
    }
  }

  public String generateInstProtocol(DetectedConnectors dc) throws BadBindException, CDLExceptLock, CDLExceptRemote, CannotGenerateException {
    StringBuffer protocol = new StringBuffer();
    boolean first = true;
    CompContained[] cont = contents();
    for(int i=0;i<cont.length;i++) {
      if (cont[i].objectKind() == ObjectsKind.o_Inst) {
        if (first) {
          first = false;
        } else {
          protocol.append("|");
        }
        protocol.append("("+ ((CompInst) cont[i]).generateArchProtocol(dc) +")");
      }
    }
    if (dc.numOfBind() > 0) {
      for(int i=0;i<dc.size();i++) {
        DetectedConnector cn = dc.getConn(i);
        if (cn.bindMode == CompBind.m_bind) {
          protocol.insert(0,"((");
          protocol.append(")&");
          RoleDescription[] cRoles = new RoleDescription[cn.cRoles.size()];
          RoleDescription[] sRoles = new RoleDescription[cn.sRoles.size()];
          String[] operations = new String [cn.operations.length];
          String prep;
          // genrate sRole
          for (int k=0;k<cn.sRoles.size();k++) {
            operations = new String [cn.operations.length];
            CompBindOper sRole = (CompBindOper) cn.sRoles.get(k);
            if (sRole.kind == CompBindOper.b_provreq) {
              sRole.elements.toFirst();
              prep = "<sRole"+Integer.toString(k)+"-"+((CompBindType)sRole.elements.aktual()).name+">";
            } else {
              sRole.subs.toFirst();
              sRole.elements.toFirst();
              prep = "<sRole"+Integer.toString(k)+"-"+((CompBindType)sRole.subs.aktual()).name+":"+((CompBindType)sRole.elements.aktual()).name+">";
            }
            for (int j=0; j<operations.length;j++) {
              operations[j] = prep +"."+ cn.operations[j];
            }
            sRoles[k] = new RoleDescription(sRole.toString(), operations);
          }
          //generate cRoles
          for (int k=0;k<cn.cRoles.size();k++) {
            operations = new String [cn.operations.length];
            CompBindOper cRole = (CompBindOper) cn.cRoles.get(k);
            if (cRole.kind == CompBindOper.b_provreq) {
              cRole.elements.toFirst();
              prep = "<"+((CompBindType)cRole.elements.aktual()).name+"-cRole"+Integer.toString(k)+">";
            } else {
              cRole.subs.toFirst();
              cRole.elements.toFirst();
              prep = "<"+((CompBindType)cRole.subs.aktual()).name+":"+((CompBindType)cRole.elements.aktual()).name+"-cRole"+Integer.toString(k)+">";
            }
            for (int j=0; j<operations.length;j++) {
              operations[j] = prep +"."+ cn.operations[j];
            }
            cRoles[k] = new RoleDescription(cRole.toString(), operations);
          }
          // roles generated
          
          for (int j=0; j<cRoles.length; j++) {
            for (int k=0; k<operations.length; k++) {
              protocol.append(cRoles[j].methods[k]);
              protocol.append(",");
            }
          }
          for (int k=0; k<operations.length; k++) {
            if (k!=0)
              protocol.append(",");
            protocol.append(sRoles[0].methods[k]);
          }
          protocol.append("&");

          SOFA.SOFAnode.Util.ConnectorProtocolGen.ConnectorProtocolGen generator;
          try {
            generator=SOFA.SOFAnode.Util.ConnectorProtocolGen.GenFactory.getConnectorProtocolGen(cn.using);
          } catch (Exception e) {
            throw new BadBindException("Unknown connector type '"+cn.using+"'",e);
          }
          protocol.append("("+generator.generate(cRoles, sRoles)+")");
          protocol.append(")");
        }
      }
    }
    
    return protocol.toString();
   /* arrayOfArraysOfInsts aai = new arrayOfArraysOfInsts();
    CompContained[] cont = contents();
    for (int i=0;i<cont.length;i++) {
      if (cont[i].objectKind() == ObjectsKind.o_Inst) {
        arrayOfInsts ai = new arrayOfInsts();
        ai.add(new protInst((CompInst)cont[i]));
        aai.add(ai);
      }
      if (cont[i].objectKind() == ObjectsKind.o_Bind) {
        CompBind bind = (CompBind) cont[i];
        if (bind.mode == CompBind.m_bind && bind.lhs.kind==CompBindOper.b_sub) {
          int ri, li;  // index of rhs, index of lhs
          bind.rhs.subs.toFirst(); bind.lhs.subs.toFirst();
          bind.rhs.elements.toFirst(); bind.lhs.elements.toFirst();
          ri = aai.containsName(((CompBindType)bind.rhs.subs.aktual()).name);
          li = aai.containsName(((CompBindType)bind.lhs.subs.aktual()).name);
          if (ri == li) {  // this subcomps are already connected
            arrayOfInsts rarr = aai.myGet(ri);  // get current list
            ri = rarr.containsName(((CompBindType)bind.rhs.subs.aktual()).name);
            li = rarr.containsName(((CompBindType)bind.lhs.subs.aktual()).name);
            int max = Math.max(ri, li);
            protInst cur = rarr.myGet(ri);  // provide inst
            String provName = ((CompBindType) bind.rhs.elements.aktual()).name;
            int j;
            for (j=0;j<cur.inst.provOps.size();j++) {  // searching provision in Inst
              if ( ((CompInst.provIfaceOperations)cur.inst.provOps.get(j)).provide.compareTo(provName)==0 )
                break;
            }
            protInst toAdd = rarr.myGet(max);
            toAdd.addToSet(((CompInst.provIfaceOperations)cur.inst.provOps.get(j)).toRewOperations());
          } else { // connect these subcomps to one
            arrayOfInsts rarr = (arrayOfInsts)aai.remove(ri);
            protInst cur = rarr.myGet(0);  // first inst
            String provName = ((CompBindType) bind.rhs.elements.aktual()).name;
            int j;
            boolean found = false;
            for (j=0;j<cur.inst.provOps.size();j++) {  // searching provision in Inst
              if ( ((CompInst.provIfaceOperations)cur.inst.provOps.get(j)).provide.compareTo(provName)==0 ) {
                found = true;
                break;
              }
            }
            if (!found) {  // some inst is used in several binds
              throw new BadBindException();
            }
            cur.addToSet(((CompInst.provIfaceOperations)cur.inst.provOps.get(j)).toRewOperations());
            // removed "ri" --> get new index of "li"
            li = aai.containsName(((CompBindType)bind.lhs.subs.aktual()).name);
            aai.myGet(li).appendAI(rarr);
          }
        }
      }
    }
    return aai.protocol();

    */
  }

  public void createOpArrays() throws CDLExceptLock, CDLExceptRemote{
    int i,j;
    java.util.ArrayList prov = new java.util.ArrayList();
    java.util.ArrayList all = new java.util.ArrayList();
    if (frame.isNew) {  // frame is currently compiled
      CompFrame fr = (CompFrame) frame.getRefferedObject(repository());
      CompContained[] cont = fr.contents();
      for (i=0;i<cont.length;i++) {
        if (cont[i].objectKind() == ObjectsKind.o_Provides) {
          String[] ops = ((CompProvide)cont[i]).operations.toArray();
          String prep="";
          for(j=0;j<provReq.size();j++) {
            if ( ((reProvReq)provReq.get(j)).orig.compareTo(((CompProvide)cont[i]).getIdent().name)==0)
              prep = ((reProvReq)provReq.get(j)).rew;
          }
          CompFrame.prependStringArray("?"+prep+".",ops);
          for(j=0;j<ops.length;j++) {
            all.add(ops[j]);
            prov.add(ops[j]);
          }
        }
        if (cont[i].objectKind() == ObjectsKind.o_Requires) {
          String[] ops = ((CompRequire)cont[i]).operations.toArray();
          String prep="";
          for(j=0;j<provReq.size();j++) {
            if ( ((reProvReq)provReq.get(j)).orig.compareTo(((CompRequire)cont[i]).getIdent().name)==0)
              prep = ((reProvReq)provReq.get(j)).rew;
          }
          CompFrame.prependStringArray("!"+prep+".",ops);
          for(j=0;j<ops.length;j++) {
            all.add(ops[j]);
          }
        }
      }
    } else {  // frame is in repository
      try {
        FrameDef fr = (FrameDef) frame.getRefferedObject(repository());
        Contained[] cont = fr.contents(null);
        for (i=0;i<cont.length;i++) {
          if (cont[i].get_def_kind().value() == DefinitionKind.dk_Provides) {
            String[] ops = CompInst.getNormFrameProvideOperations((ProvideDef)cont[i], repository());
            String prep="";
            for(j=0;j<provReq.size();j++) {
              if ( ((reProvReq)provReq.get(j)).orig.compareTo(((ProvideDef)cont[i]).get_identification().name())==0)
                prep = ((reProvReq)provReq.get(j)).rew;
            }
            CompFrame.prependStringArray("?"+prep+".",ops);
            for(j=0;j<ops.length;j++) {
              prov.add(ops[j]);
              all.add(ops[j]);
            }
          }
          if (cont[i].get_def_kind().value() == DefinitionKind.dk_Requires) {
            String[] ops = CompInst.getNormFrameRequireOperations((RequireDef)cont[i], repository());
            String prep="";
            for(j=0;j<provReq.size();j++) {
              if ( ((reProvReq)provReq.get(j)).orig.compareTo(((RequireDef)cont[i]).get_identification().name())==0)
                prep = ((reProvReq)provReq.get(j)).rew;
            }
            CompFrame.prependStringArray("!"+prep+".",ops);
            for(j=0;j<ops.length;j++) {
              all.add(ops[j]);
            }
          }
        }
      } catch (RemoteException e) {
        throw new CDLExceptRemote(e.getMessage());
      } catch (TIRExceptLock e) {
        throw new CDLExceptLock(e.getMessage());
      }
    }
    provOpArray = new String[prov.size()];
    allOpArray = new String[all.size()];
    for(i=0;i<prov.size();i++) {
      provOpArray[i] = (String) prov.get(i);
    }
    for(i=0;i<all.size();i++) {
      allOpArray[i] = (String) all.get(i);
    }
  }

  // *************** new protocols (with connectors) *************
  

  /** Perform several tests, if bindings are correct and detects, which binds create connector.
    * * Currently tests, if one interface is binded by connectors of one kind.
    * @exception BadBindException     bad bind occured
    * @return   detected connectors
    */
  private DetectedConnectors detectConnectors() throws BadBindException, CDLExceptRemote, CDLExceptLock {
    DetectedConnectors dc = new DetectedConnectors();
    CompContained[] cont = contents();
    CompBindOper rhs, lhs;
    for (int i=0;i<cont.length;i++) {
      if (cont[i].objectKind() == ObjectsKind.o_Bind) {
        try {
          SOFA.SOFAnode.Util.ConnectorProtocolGen.GenFactory.addDetectedConnector(getConnectorOperations((CompBind)cont[i]),dc,(CompBind)cont[i]);
        } catch (GeneratorNotFoundException e) {
          throw new BadBindException("Unknown connector type used for binding.",e);
        }
      }
    }
    return dc;
  }  // *** end of detectConnectors()


  private String[] getConnectorOperations(CompBind bind) throws BadBindException, CDLExceptLock, CDLExceptRemote {
    int mode = bind.mode;
    if (mode == CompBind.m_subsume || mode == CompBind.m_delegate || (mode == CompBind.m_bind &&bind.rhs.kind==CompBindOper.b_provreq && bind.lhs.kind==CompBindOper.b_provreq ) ) {  // this frame
      String sname;
      if (mode == CompBind.m_delegate) {
        bind.lhs.elements.toFirst();
        sname = ((CompBindType) bind.lhs.elements.aktual()).name;
      } else {
        bind.rhs.elements.toFirst();
        sname = ((CompBindType) bind.rhs.elements.aktual()).name;
      }
      if (frame.isNew) {  // frame is currently compiled
        CompFrame fr = (CompFrame) frame.getRefferedObject(repository());
        CompContained[] prq = fr.lookup(sname);
        if (prq==null || prq.length == 0)
          throw new BadBindException("Unknown part \""+sname+"\" of bind in "+id.toString());
        if (prq[0].objectKind() == ObjectsKind.o_Provides) 
          return ((CompProvide)prq[0]).operations.toArray();
        if (prq[0].objectKind() == ObjectsKind.o_Requires) 
          return ((CompRequire)prq[0]).operations.toArray();
      } else {  // frame is in repository
        try {
          FrameDef fr = (FrameDef) frame.getRefferedObject(repository());
          Contained[] prq = fr.lookup_name(sname);
          if (prq==null || prq.length == 0)
            throw new BadBindException("Unknown part \""+sname+"\" of bind in "+id.toString());
          if (prq[0].get_def_kind().value() == DefinitionKind.dk_Provides) 
            return CompInst.getNormFrameProvideOperations((ProvideDef)prq[0], repository());
          if (prq[0].get_def_kind().value() == DefinitionKind.dk_Requires) 
            return CompInst.getNormFrameRequireOperations((RequireDef)prq[0], repository());
        } catch (RemoteException e) {
          throw new CDLExceptRemote(e.getMessage());
        } catch (TIRExceptLock e) {
          throw new CDLExceptLock(e.getMessage());
        }
      }
    } else {  // it is bind between instances
      String sname;
      String iname;
      bind.rhs.elements.toFirst();
      bind.rhs.subs.toFirst();
      sname = ((CompBindType) bind.rhs.elements.aktual()).name;
      iname = ((CompBindType) bind.rhs.subs.aktual()).name;
      CompContained[] inst = lookup(iname);
      if (inst==null || inst.length == 0)
        throw new BadBindException("Unknown part \""+iname+"\" of bind in "+id.toString());
      if (((CompInst)inst[0]).subcom.isNew) {
        CompFrame fr = (CompFrame) CompInst.getInstFrame(((CompInst)inst[0]).subcom, repository()).getRefferedObject(repository());
        CompContained[] prq = fr.lookup(sname);
        if (prq==null || prq.length == 0)
          throw new BadBindException("Unknown part \""+sname+"\" of bind in "+id.toString()+"\n ### "+ fr.id.toString());
        if (prq[0].objectKind() == ObjectsKind.o_Provides) 
          return ((CompProvide)prq[0]).operations.toArray();
        if (prq[0].objectKind() == ObjectsKind.o_Requires) 
          return ((CompRequire)prq[0]).operations.toArray();
      } else {
        try {
          FrameDef fr = (FrameDef) CompInst.getInstFrame(((CompInst)inst[0]).subcom, repository()).getRefferedObject(repository());
          Contained[] prq = fr.lookup_name(sname);
          if (prq==null || prq.length == 0)
            throw new BadBindException("Unknown part \""+sname+"\"of bind in "+id.toString());
          if (prq[0].get_def_kind().value() == DefinitionKind.dk_Provides) 
            return CompInst.getNormFrameProvideOperations((ProvideDef)prq[0], repository());
          if (prq[0].get_def_kind().value() == DefinitionKind.dk_Requires) 
            return CompInst.getNormFrameRequireOperations((RequireDef)prq[0], repository());
        } catch (RemoteException e) {
          throw new CDLExceptRemote(e.getMessage());
        } catch (TIRExceptLock e) {
          throw new CDLExceptLock(e.getMessage());
        }
      }
    }
  

    return null;
  }

}
