/* $Id: ContainedGeneratedObject.java,v 1.1.1.1 2003/02/11 16:19:40 bures Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

import java.io.IOException;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.RemoteException;
 
/** Representation of already generated contained object. It's for generating of correct references to 
  * this object, especially for generating reference to typedefs.
  * 
  * @author Petr Hnetynka
  */
public class ContainedGeneratedObject extends GeneratedObject {

  /** object in TIR */
  Contained obj;
  /** generated name */
  String generatedName;

  String inRef;
  String outRef;
 
  /** Constructor
    *
    * @param obj   represented repository object
    * @param generatedName  name of generated type
    * @exception RemoteException  remote exception
    */
  public ContainedGeneratedObject(Contained obj, String generatedName) throws RemoteException {
    this.obj = obj;
    this.generatedName = generatedName;
    key = createKey(obj);
    inRef = generatedName;
    outRef = generatedName+"Holder";
  }

/** Constructor
    *
    * @param obj   represented repository object
    * @param inRefer   in reference
    * @param outRefer  out reference
    * @exception RemoteException  remote exception
    */
  public ContainedGeneratedObject(Contained obj, String inRefer, String outRefer) throws RemoteException {
    this.obj = obj;
    this.generatedName = inRefer;
    key = createKey(obj);
    inRef = inRefer;
    outRef = outRefer;
  }
  
  /** Compare identification of this object with given absolute name and version.
    *
    * @param  absName  absolute name of object
    * @param  ver      version identifier
    * @return <tt>true</tt> if this is object with given identification
    * @exception RemoteException     remote exception
    */
  public boolean isEqual(String absName, String ver) throws RemoteException {
    Identification idl = obj.get_identification();
    return absName.compareTo(idl.absolute_name().name())==0 && ver.compareTo(idl.version())==0;
  }

  /** Compare identification of this object with given identification.
    *
    * @param  identification      tested identification
    * @return <tt>true</tt>   if this is object with given identification
    * @exception RemoteException     remote exception
    */
  public boolean isEqual(Identification identification)  throws RemoteException {
    return obj.get_identification().is_equal(identification);
  }

  // inherited from GeneratedObject
  public String genReference(int kind) {
    switch (kind) {
    case INREF:  return inRef;
    case OUTREF: return outRef;
    default:     return null;
    }
  }

  // inherited from GeneratedObject
  public void genReference(CGFileWriter out, int kind) throws IOException {
    switch (kind) {
    case INREF:  out.iwrite(inRef);
    case OUTREF: out.iwrite(outRef);
    default:     ;
    }
  }

  // inherited from GeneratedObject
  public String genSpecReference(int kind, int specType) {
    if (specType == 0) {
      switch (kind) {
      case INREF:  return inRef;
      case OUTREF: return outRef;
      default:     return null;
      }
    } else if (specType == 1) {  // CN interface
      int n = inRef.lastIndexOf('.');
      StringBuffer sb = new StringBuffer(inRef);
      if (n != -1) {
        sb.insert(n+1,"CN");
      } else {
        sb.insert(inRef.lastIndexOf(":")+1,"CN");
      }
      return sb.toString();
    }
    return null;
  }

  // inherited from GeneratedObject
  public void genSpecReference(CGFileWriter out, int kind, int specType) throws IOException {
    out.iwrite(genSpecReference(kind,specType));
  }


  public void defaultValue(CGFileWriter out) throws IOException {
    out.iwriteln("null");
  }

  public String defaultValue() throws IOException {
    return "null";
  } 

  public void genReadType(CGFileWriter out, String varName) throws IOException {
    out.iwrite(varName+" = "+generatedName+"Helper.read("+defaultIstream+")");
  }

  public String genReadType(String varName) throws IOException {
    return varName+" = "+generatedName+"Helper.read("+defaultIstream+")";
  }

  public void genWriteType(CGFileWriter out, String varName) throws IOException {
    out.iwrite(generatedName+"Helper.write("+defaultOstream+", "+ varName+")");
  }

  public String genWriteType(String varName) throws IOException {
    return generatedName+"Helper.write("+defaultOstream+", "+ varName+")";
  }

  public String toString() {
    return inRef;
  }
}
