/* $Id: GenObjectList.java,v 1.1.1.1 2003/02/11 16:19:40 bures Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

/** Collection of allready enerated objects.
  *
  * @author Petr Hnetynka
  */
public class GenObjectList extends java.util.ArrayList {

  public static class ExistException extends CodeGenException {
    public ExistException() { super(); };
    public ExistException(String s) { super(s); };
  }

  /** Returns the object at the specified position in this list.
    * 
    * @exception IndexOutOfBoundsException     if index out of range (index < 0 || index >= size()).
    */
  public GeneratedObject getObject(int index) {
    return (GeneratedObject) get(index);
  }
 
  /** Returns the index of object with specified key.
    *
    * @param key    key of the searched object
    * @return the index of the the argument in this list; returns -1 if the object is not found. 
    */
  public int indexOfObject(String key) {
    for (int i=0; i<size(); i++) {
      if (getObject(i).compareKey(key))
        return i;
    }
    return -1;
  }

  /** Returns the object with specified key.
    *
    * @param key    key of the searched object
    * @return the searched object; returns <B>null</B> if the object is not found. 
    */
  public GeneratedObject searchObject(String key) {
    for (int i=0; i<size(); i++) {
      if (getObject(i).compareKey(key))
        return getObject(i);
    }
    return null;
  }

  /** Returns <tt>true</tt> if this list contains the specified object.
    *
    * @param key    key of the searched object
    */
  public boolean containsObject(String key) {
    for (int i=0;i<size();i++) {
      if (getObject(i).compareKey(key))
        return true;
    }
    return false;
  }

  /** Adds object to list only if this object already is not in list (uses containsObject()), else
    * it throws exception.
    *
    * @param   obj   added object 
    * @exception ExistException  object is allready in list 
    */
  public void addObject(GeneratedObject obj) throws ExistException {
    if (!containsObject(obj.key))
      add(obj);
    else
      throw new ExistException(Res.res.getString("MSG_ObjIsInList"));
  }
}
