/* $Id: IDLMapping.java,v 1.1.1.1 2003/02/11 16:19:40 bures Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

import SOFA.SOFAnode.Made.TIR.*;

/** Mapping of CDL types to IDL.
  *
  * @author Petr Hnetynka
  */
public class IDLMapping implements Mapping {
  /** Returns name of the type in IDL.
    * @param type    cdl type (TIR object)
    * @return   string with type name or null
    */
  public String getImplTypeName(CDLType type) {
    try {
      if (type instanceof PrimitiveDef) {
        int primKind = ((PrimitiveDef) type).kind().value();
        switch (primKind) {
          case PrimitiveKind.pk_void:
            return "void";
          case PrimitiveKind.pk_ushort:
            return "unsigned short";
          case PrimitiveKind.pk_short:
            return "short";
          case PrimitiveKind.pk_ulong:
            return "unsigned long";
          case PrimitiveKind.pk_long:
            return "long";
          case PrimitiveKind.pk_ulonglong:
            return "unsigned long long";
          case PrimitiveKind.pk_longlong:
            return "long long";
          case PrimitiveKind.pk_float:
            return "float";
          case PrimitiveKind.pk_double:
            return "double";
          case PrimitiveKind.pk_longdouble:
            return "long double";
          case PrimitiveKind.pk_boolean:
            return "boolean";
          case PrimitiveKind.pk_wchar:
            return "wchar";
          case PrimitiveKind.pk_char:
            return "char";
          case PrimitiveKind.pk_octet:
            return "octet";
          case PrimitiveKind.pk_wstring:
            return "wstring";
          case PrimitiveKind.pk_string:
            return "string";
          case PrimitiveKind.pk_any:
            return "any";
          case PrimitiveKind.pk_object:
            return "object";
        }
      } else {
        StringBuffer sb = new StringBuffer();
        Contained cnt = (Contained) type;
        while (! (cnt instanceof CDLRepository)) {
          Identification id = cnt.get_identification();
          String version = id.version();
          String name = id.name();
          if (sb.length() == 0) {
            sb.append(name);
          } else {
            sb.insert(0,".");
            sb.insert(0,name);
          }
          cnt = (Contained) cnt.get_defined_in(); 
        }
        return sb.toString();
      }
    } catch (java.rmi.RemoteException e) {
      return null;
    }
    return null;
  }
}
