/* $Id: BindTypeArrayImpl.java,v 1.1.1.1 2003/02/11 16:19:41 bures Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;

public class BindTypeArrayImpl extends UnicastRemoteObject implements BindTypeArray, TIRImplObject {
  DefinitionKindImpl defKindImpl;
  StateKindImpl stKindImpl;
  BindTypeKind btk;
  String nm;
  ExprOperDef[] inds;
  
  public BindTypeArrayImpl(String name, ExprOperDef[] indexes) throws RemoteException {
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_BindType);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    btk = new BindTypeKindImpl(BindTypeKind.btk_array);
    nm = name;
    inds = indexes;
  }

  public BindTypeArrayImpl() throws RemoteException {
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_BindType);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    btk = new BindTypeKindImpl(BindTypeKind.btk_array);
    nm = null;
    inds = null;
  }

/* from interface TIRObject */
  public DefinitionKind get_def_kind() throws RemoteException {
    return (DefinitionKind) defKindImpl;
  }

  /* from interface TIRObject */
  public StateKind get_state() throws RemoteException {
    return (StateKind) stKindImpl;
  }

  public BindTypeKind get_bt_kind() throws RemoteException {
    return btk;
  }

  public String name() throws RemoteException {
    return nm;
  }

  public ExprOperDef[] array_indexes() throws RemoteException {
    return inds;
  }

  public void save(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      st.curOutFile.writeInt(btk.value());
      Storage.writeString(st.curOutFile, nm);
      st.curOutFile.writeInt(inds.length);
      for(int i=0;i<inds.length;i++) 
        Storage.writeExprOperDef(st, inds[i]);
    } catch (IOException e) {
      throw new TIRExceptStorage("Can't write in directory "+st.current+".");
    }   
  }

  public void load(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      nm = Storage.readString(st.curInFile);
      int len = st.curInFile.readInt();
      inds = new ExprOperDef [len];
      for(int i=0;i<len;i++)
        inds[i] = Storage.readExprOperDef(st);
    } catch (IOException e) {
      throw new TIRExceptStorage("Can't read in directory "+st.current+".");
    }
  }

  public void postLoad(RepositoryImpl r) throws RemoteException, TIRExceptStorage {
    for(int i=0;i<inds.length;i++)
      ((TIRImplObject) inds[i]).postLoad(r);
  }

  public boolean isNew() {
    return (stKindImpl.value()==StateKind.sk_work);
  }

  public void canCommit() throws RemoteException, TIRExceptCommit {
    for(int i=0;i<inds.length;i++)
      ((TIRImplObject) inds[i]).canCommit();
  }

  public void doCommit(Container in, Repository rep) throws RemoteException {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    if (isNew()) {
      for(int i=0;i<inds.length;i++) {
        inds[i] = ExprHelp.toNormal(inds[i]);
        ((TIRImplObject) inds[i]).doCommit(in,rep);
      }
      stKindImpl.toNormal();
    }
  }

  public void doAbort(long workId) throws RemoteException {
    for(int i=0;i<inds.length;i++)
      ((TIRImplObject) inds[i]).doAbort(workId);
  }
}
