/* $Id: IdentificationImpl.java,v 1.1.1.1 2003/02/11 16:19:41 bures Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;

public class IdentificationImpl extends UnicastRemoteObject implements Identification, SIdentification {
  public String nm;    // name
  public String ver;   // version
  public String[] tg;  // tag
  public String brtg;  // branchTag
  public AbsoluteNameImpl abs;
  public AbsoluteNameImpl langabs;

  public String   lang;   // identification of description language

  Lock nextLock;       // lock for getting next version
  long lockId;         // id of locking "transaction" 
  long nextNBranch;    // number of next branch
  boolean hasNext;     // has new version

  // name must be absolute name (::aaa::bbb:ccc)
  public IdentificationImpl(String lng, String name, String version) throws RemoteException {
    super();
    lang = new String(lng);
    abs = new AbsoluteNameImpl(name);
    langabs = new AbsoluteNameImpl(lng, abs);
    if (abs.size() != 0) {
      nm = new String(abs.elementAt(abs.size()-1));
    } else { 
      nm = "";
    }
    ver = new String(version); 
    brtg = ""; tg = new String [0];
    nextLock = new Lock(false);
    lockId = -1L;
    hasNext = false;
    nextNBranch = 1;
  }

  // name must be absolute name (::aaa::bbb:ccc)
  public IdentificationImpl(String lng, String name, String version, String branchTag) throws RemoteException {
    super();
    lang = new String(lng);
    abs = new AbsoluteNameImpl(name);
    langabs = new AbsoluteNameImpl(lng, abs);
    if (abs.size() != 0) {
      nm = new String(abs.elementAt(abs.size()-1));
    } else { 
      nm = "";
    }
    ver = new String(version); 
    brtg = new String(branchTag); tg = new String [0];
    nextLock = new Lock(false);
    lockId = -1L;
    hasNext = false;
    nextNBranch = 1;
  }

  public IdentificationImpl() throws RemoteException {
    nm = "";
    ver = "";
    lang = "";
    tg = new String [0];
    brtg = "";
    abs = new AbsoluteNameImpl("");
    langabs = null;
    nextLock = new Lock(false);
    lockId = -1L;
    hasNext = false;
    nextNBranch = 1;
  }  

  public IdentificationImpl(Identification i) throws RemoteException {
    nm = new String(i.name());
    ver = new String(i.version());
    brtg = new String(i.branchTag());
    String[] t = i.tag();
    int j;
    tg = new String [t.length];
    for(j=0;j<t.length;j++)
      tg[j] = new String(t[j]);
    abs = new AbsoluteNameImpl(i.absolute_name().name());
    nextLock = new Lock(false);
    lockId = -1L;
    hasNext = ((SIdentification)i).hasNextVersion();
    nextNBranch = ((SIdentification)i).nextNumBranch();

    lang = new String(((SIdentification)i).language());
    
    langabs = new AbsoluteNameImpl(lang, abs); 
  }

  public String language() throws RemoteException {
    return lang;
  }

  public boolean is_equal(Identification id) throws RemoteException {
    if ((nm.compareTo(id.name())==0) && (ver.compareTo(id.version())==0)
        && abs.isEqual(id.absolute_name()) && (lang.compareTo(((SIdentification)id).language())==0))
      return true;
    else 
      return false;
  }
  
  public boolean is_name_equal(Identification id) throws RemoteException {
    if ((nm.compareTo(id.name())==0)&& abs.isEqual(id.absolute_name()))
      return true;
    else 
      return false;
  }

  public boolean is_name_equal(String name) throws RemoteException {
    if (name.compareTo(abs.name())==0)
      return true;
    else 
      return false;
  }

  public String name() {
    return nm;
  }

  public AbsoluteName absolute_name() { 
    return abs;
  }

  public AbsoluteName lang_absolute_name() throws RemoteException { 
    return langabs;
  }

  public String version() {
    return ver;
  }

  public void version(String v) {
    if (v==null)
      ver = "";
    ver = v;
  }

  public void tag(String t) {
    if (t!=null) {
      int i, len = tg.length;
      for (i=0;i<len;i++) {
        if (tg[i].compareTo(t)==0)
          return;
      }
      String[] newt = new String [len+1];
      for (i=0;i<len;i++) {
        newt[i] = tg[i];
      }
      newt[i] = new String(t);
      tg = newt;
    }
  }

  public String[] tag() {
    String t[] = new String [tg.length];
    for (int i=0;i<tg.length;i++)
      t[i] = new String(tg[i]);
    return t;
  }

  public String branchTag() {
    return brtg;
  }

  public void save(DataOutputStream out) throws TIRExceptStorage {
    try {
      Storage.writeString(out, nm);
      Storage.writeString(out, abs.toString());
      Storage.writeString(out, lang);
      Storage.writeString(out, ver);
      out.writeInt(tg.length);
      for (int i=0; i<tg.length;i++)
        Storage.writeString(out, tg[i]);
      Storage.writeString(out, brtg);
      out.writeLong(nextNBranch);
      if (nextLock.isLocked())
        out.writeBoolean(false);
      else  
        out.writeBoolean(hasNext);
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in saving identification "+nm);
    }
  }

  public void load(DataInputStream in) throws TIRExceptStorage {
    try {
      nm = Storage.readString(in);
      abs = new AbsoluteNameImpl(Storage.readString(in));
      lang = Storage.readString(in);
      langabs = new AbsoluteNameImpl(lang, abs);
      ver = Storage.readString(in);
      int i, len = in.readInt();
      tg = new String [len];
      for(i=0;i<len;i++) 
        tg[i] = Storage.readString(in);
      brtg = Storage.readString(in);
      nextNBranch = in.readLong();
      hasNext = in.readBoolean();
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error");
    } 
  }

  public boolean is_equal(String name, String version) throws RemoteException {
    if (name.compareTo(abs.name())==0 && (ver.compareTo(version)==0))
      return true;
    else 
      return false;
  }

  public String toString() {
    return new String(nm+"?"+ver);
  }

  public boolean isLocked() throws RemoteException {
    return nextLock.isLocked();
  }

  public long lockFor() throws RemoteException {
    return lockId;
  }
  
  public boolean lock(long lockFor) throws RemoteException {
    boolean a = nextLock.lock();
    if (a)
      this.lockId = lockFor;
    return a;
  }
  
  public void unlock() throws RemoteException {
    lockId = -1L;
    nextLock.unlock();
  }
  
  public long nextNumBranch() throws RemoteException {
    return nextNBranch;
  }

  public void incNextNumBranch() throws RemoteException { nextNBranch++; }
  public boolean hasNextVersion() throws RemoteException { return hasNext; }
  public void setNextVersion() throws RemoteException { hasNext = true; }

  public boolean is_short_equal(Identification id) throws RemoteException {
    if ((nm.compareTo(id.name())==0) && (ver.compareTo(id.version())==0))
      return true;
    else 
      return false;
  }
  
  public boolean is_short_equal(String name, String version) throws RemoteException {
    if ((nm.compareTo(name)==0) && (ver.compareTo(version)==0))
      return true;
    else 
      return false;
  }
  
  public boolean is_short_name_equal(Identification id) throws RemoteException {
    if ((nm.compareTo(id.name())==0))
      return true;
    else 
      return false;
  }

  public boolean is_short_name_equal(String name) throws RemoteException {
    if (nm.compareTo(name)==0)
      return true;
    else 
      return false;
  }

  public boolean is_local() throws RemoteException {
    String sofaNodeName = System.getProperty("sofanode.name","");

    //test same node 
    StringBuffer ip = new StringBuffer();
    int i = 0;
    while ((i<ver.length()) && (ver.charAt(i)!='!')) {
      ip.append(ver.charAt(i));
      i++;
    }
  
    if (sofaNodeName.compareTo(ip.toString())==0)
      return true;
    else
      return false;
  }
}
