/* $Id: ProtocolBinOperationDefImpl.java,v 1.1.1.1 2003/02/11 16:19:41 bures Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.IOException;

public class ProtocolBinOperationDefImpl extends UnicastRemoteObject implements ProtocolBinOperationDef, TIRImplObject, SProtocolOperDef {
  ProtocolOperKindImpl pok;
  DefinitionKindImpl defKindImpl;
  StateKindImpl stKindImpl;
  ProtocolOperDef oper1;
  ProtocolOperDef oper2;

  public ProtocolBinOperationDefImpl(int kind, ProtocolOperDef op1, ProtocolOperDef op2) throws RemoteException {
    pok = new ProtocolOperKindImpl(kind);
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_ProtocolOper);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    oper1 = op1;
    oper2 = op2;
  }

  public ProtocolBinOperationDefImpl(int kind) throws RemoteException {
    pok = new ProtocolOperKindImpl(kind);
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_ProtocolOper);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    oper1 = null;
    oper2 = null;
  }

  public ProtocolBinOperationDefImpl() throws RemoteException {
    pok = null;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Protocol);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    oper1 = null;
    oper2 = null;
  }

  /* from interface TIRObject */
  public DefinitionKind get_def_kind() throws RemoteException {
    return (DefinitionKind) defKindImpl;
  }

  /* from interface TIRObject */
  public StateKind get_state() throws RemoteException {
    return (StateKind) stKindImpl;
  }

  public ProtocolOperDef operand1() throws RemoteException {
    return oper1;
  }

  public ProtocolOperDef operand2() throws RemoteException {
    return oper2;
  }

  public ProtocolOperKind get_po_kind() throws RemoteException {
    return pok;
  }

  public void save(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      st.curOutFile.writeInt(pok.value());
      Storage.writeProtocolOperDef(st, oper1);
      Storage.writeProtocolOperDef(st, oper2);
    } catch (IOException e) {
      throw new TIRExceptStorage("Can't write in directory "+st.current+".");
    }
  }

  public void load(Storage st) throws RemoteException, TIRExceptStorage {
    oper1 = Storage.readProtocolOperDef(st);
    oper2 = Storage.readProtocolOperDef(st);
  }

  public void postLoad(RepositoryImpl r) throws RemoteException, TIRExceptStorage {
    ((TIRImplObject) oper1).postLoad(r);
    ((TIRImplObject) oper2).postLoad(r);
  }

  public boolean isNew() {
    return (stKindImpl.value()==StateKind.sk_work);
  }

  public void canCommit() throws RemoteException, TIRExceptCommit {;}

  public void doCommit(Container in, Repository rep) throws RemoteException {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    stKindImpl.toNormal();
    ((TIRImplObject) oper1).doCommit(in,rep);
    ((TIRImplObject) oper2).doCommit(in,rep);
  }

  public void doAbort(long workId) throws RemoteException {}

  public String to_text() throws RemoteException {
    StringBuffer sb = new StringBuffer("(");
    String strOp="";
    switch (pok.value()) {
    case ProtocolOperKind.pok_alternative:
      strOp = ")+(";
      break;
    case ProtocolOperKind.pok_andparallel:
      strOp = ")|(";
      break;
    case ProtocolOperKind.pok_orparallel:
      strOp = ")||(";
      break;
    case ProtocolOperKind.pok_sequence:
      strOp = ");(";
      break;
    case ProtocolOperKind.pok_restriction:
      strOp = ")/(";
      break;
    default: break;  
    }
    
    sb.append(((SProtocolOperDef)oper1).to_text());
    sb.append(strOp);
    sb.append(((SProtocolOperDef)oper2).to_text());
    sb.append(")");
    return sb.toString();
  }
}
