/* $Id: StructMemberImpl.java,v 1.1.1.1 2003/02/11 16:19:41 bures Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;

public class StructMemberImpl extends ContainedImpl implements StructMember, TIRImplObject {
  String nm;
  CDLType tp;
  DefinitionKindImpl defKindImpl;
  StateKindImpl stKindImpl;

  StructMemberImpl normal;

  public StructMemberImpl(Identification ident, Container in, Repository inrep, CDLType type) throws RemoteException {
    super(ident,in,inrep);
    tp = type;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_StructMember);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    normal = null;
  }

  /* constuctor for load() */
  public StructMemberImpl(Container in, Repository inrep) throws RemoteException {
    super(null,in,inrep);
    tp = null;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_StructMember);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    normal = null;
  } 

  public StructMemberImpl(StructMemberImpl a, Container in, Repository inrep) throws RemoteException {
    super(a.id,in,inrep);
    tp = a.tp;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_StructMember);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    normal = a;
  }

  public String name() throws RemoteException {
    return id.name();
  }
  
  public CDLType type() throws RemoteException {
    return tp;
  }

  /* from interface TIRObject */
  public DefinitionKind get_def_kind() throws RemoteException {
    return (DefinitionKind) defKindImpl;
  }

  /* from interface TIRObject */
  public StateKind get_state() throws RemoteException {
    return (StateKind) stKindImpl;
  }

  public void tag(String t) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (stKindImpl.value()!=StateKind.sk_work)
      throw new TIRExceptCreate("you can call create method on work object");
    super.tag(t);
  }

  public void load(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      id = new IdentificationImpl();
      ((IdentificationImpl)id).load(st.curInFile);
      tp = Storage.readCDLType(st);
    } catch (IOException e) {
      throw new TIRExceptStorage("Can't read in file "+st.currentFile+".");
    }
  }

  public void save(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      ((IdentificationImpl)id).save(st.curOutFile);
      Storage.writeCDLType(st,tp);
    } catch (IOException e) {
      throw new TIRExceptStorage("Can't write in file "+st.currentFile+".");
    }
  }

  public void postLoad(RepositoryImpl r) throws RemoteException, TIRExceptStorage {
    try {
    
    if (tp.get_def_kind().value() == DefinitionKind.dk_none) {   // is tp NotYetInRep?
      if (((NotYetInRep) tp).defKind == DefinitionKind.dk_Primitive)
        tp = r.get_prim(((NotYetInRep) tp).primKind);
      else {     // in tp there is reference to some type
        AbsoluteName nm = ((NotYetInRep) tp).nm;
	Container con = r;  // begin search from repository
	Contained[] rt;
	Contained ret;
	IdentificationImpl id;
        for (int i=0;i<nm.size()-1;i++) {  // last element is name of object
	  id = new IdentificationImpl(nm.elementAt(0),"::"+nm.elementAt(i),"");
          rt = ((SContainer) con).slookup_name(id);
	  if (rt.length==0) throw new TIRExceptStorage("Can't find type " + nm.name());
          if (rt.length==1) {
            con = (Container) rt[0];
	  } else {  // type in struct (etc.) - more versions of this struct
	    boolean found = false;
            id.version(((NotYetInRep) tp).id.version());  // substruct version is inherited from parent
            int j = 0;
	    for (j=0;j<rt.length;j++) {
	      if (id.is_short_equal(rt[j].get_identification())) {
	        found = true;
		break;  // -- break for
	      }
	    } // -- end of for
	    if (!found) throw new TIRExceptStorage("Can't find type " + nm.name());
	    con = (Container) rt[j];
	  }
	}
	if ((ret = ((SContainer)con).slookup(((NotYetInRep) tp).id))==null)
	  throw new TIRExceptStorage("Can't find type " + nm.name());
	tp = (CDLType) ret;
	if (tp.get_def_kind().value()==DefinitionKind.dk_Array ||
            tp.get_def_kind().value()==DefinitionKind.dk_Sequence ||
            tp.get_def_kind().value()==DefinitionKind.dk_String ||
            tp.get_def_kind().value()==DefinitionKind.dk_Wstring)
	  ((TIRImplObject) tp).postLoad(r);
      }
    } else {
      if (tp.get_def_kind().value()==DefinitionKind.dk_Array ||
          tp.get_def_kind().value()==DefinitionKind.dk_Sequence ||
          tp.get_def_kind().value()==DefinitionKind.dk_String ||
          tp.get_def_kind().value()==DefinitionKind.dk_Wstring)
        ((TIRImplObject) tp).postLoad(r);
    }

    } catch (TIRExceptLock e) {;}
  }

  public boolean isNew() {
    return ((stKindImpl.value()==StateKind.sk_work) && normal==null);
  }

  public void canCommit() throws RemoteException, TIRExceptCommit {;}

  public void doCommit(Container in, Repository rep) throws RemoteException {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    if (isNew()) {  
      stKindImpl.toNormal();
      switch (tp.get_def_kind().value()) {
        case DefinitionKind.dk_Primitive:
          tp = ((SRepository)rep).get_prim(((PrimitiveDef)tp).kind().value());
          break;
        case DefinitionKind.dk_Array:
          ArrayDef nt = new ArrayDefImpl(ExprHelp.toNormal(((ArrayDef) tp).length()), ((ArrayDef) tp).element_type());
          tp = nt;
          ((TIRImplObject)tp).doCommit(in,rep);
          break;
        case DefinitionKind.dk_Sequence:
          SequenceDef st = new SequenceDefImpl(ExprHelp.toNormal(((SequenceDef) tp).bound()), ((SequenceDef) tp).element_type());
          tp = st;
          ((TIRImplObject)tp).doCommit(in,rep);
          break;
        case DefinitionKind.dk_String:
          StringDef str = new StringDefImpl(ExprHelp.toNormal(((StringDef) tp).bound()));
          tp = str;
          ((TIRImplObject)tp).doCommit(in,rep);
          break;
        case DefinitionKind.dk_Wstring:
          WstringDef wstr = new WstringDefImpl(ExprHelp.toNormal(((WstringDef) tp).bound()));
          tp = wstr;
          ((TIRImplObject)tp).doCommit(in,rep);
          break;
        case DefinitionKind.dk_Fixed:
          FixedDef ft = new FixedDefImpl(ExprHelp.toNormal(((FixedDef) tp).digits()), ExprHelp.toNormal(((FixedDef) tp).scale()));
          tp = ft;
          ((TIRImplObject)tp).doCommit(in,rep);
          break;
        default:
          Identification idl = ((Contained)tp).get_identification();
          tp = new NotYetInRep(idl.lang_absolute_name().name(), idl.name(), idl.version());
          break;
      }
    } else {
      normal=null;
    }
  }    

  public void doAbort(long workId) throws RemoteException {}
}
