/* $Id: UnionDefImpl.java,v 1.1.1.1 2003/02/11 16:19:41 bures Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;

public class UnionDefImpl extends ContainerImpl implements UnionDef, TIRImplObject, SContained {
  protected Identification id;
  protected Container parent;
  protected Repository rep;
  CDLType swtype;
  DefinitionKindImpl defKindImpl;
  StateKindImpl stKindImpl;
 
  UnionDefImpl normal;
  long workId;
 
  public UnionDefImpl(Identification ident, Container in, Repository inrep, CDLType swType, long workId) throws RemoteException {
    id = ident;
    parent = in;
    rep = inrep;
    swtype = swType;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Union);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    normal = null;
    this.workId = workId;
  }

  public UnionDefImpl(Container in, Repository inrep) throws RemoteException {
    id = null;
    parent = in;
    rep = inrep;
    swtype = null;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Union);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    normal = null;
    this.workId = -1L;
  }

  public UnionDefImpl(UnionDefImpl a, Container in, Repository inrep, long workId) throws RemoteException {
    id = a.id;
    parent = in;
    rep = inrep;
    swtype = a.swtype;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Union);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    normal = a;
    this.workId = workId;
  }

  public CDLType switch_type() throws RemoteException {
    return swtype;
  }
 
  /* form interface Contained */
  public Identification get_identification() throws RemoteException {
    return id;
  }

  /* form interface Contained */
  public Container get_defined_in() throws RemoteException {
    return parent;
  }

  /* form interface Contained */
  public Container get_containing_repository() throws RemoteException {
    return rep;
  }

  /* form interface Contained */
  public AbsoluteName get_absolute_name() throws RemoteException {
    AbsoluteNameImpl absName;
    if (rep == parent) {  // object is contained in repository
      absName = new AbsoluteNameImpl("::"+id.name());
    } else {
      absName = new AbsoluteNameImpl(((SContained)parent).get_absolute_name().name()+"::"+id.name());
    }
    return (AbsoluteName) absName;
  }

  /* from interface TIRObject */
  public DefinitionKind get_def_kind() throws RemoteException {
    return (DefinitionKind) defKindImpl;
  }

  /* from interface TIRObject */
  public StateKind get_state() throws RemoteException {
    return (StateKind) stKindImpl;
  }

  public void tag(String t) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (stKindImpl.value()!=StateKind.sk_work)
      throw new TIRExceptCreate("you can call create method on work object");
    String ebt = id.branchTag();
    if (t.compareTo(ebt)==0)
      throw new TIRExceptCreate("This tag exists");
    String[] et = id.tag();
    int i,j;
    for (i=0;i<et.length;i++) {
      if (t.compareTo(et[i])==0)
       throw new TIRExceptCreate("This tag exists");       
    }
    Contained[] sibl = parent.lookup_name(id);
    for (j=0;j<sibl.length;j++) {
      Identification idl = sibl[j].get_identification();
      
      ebt = idl.branchTag();
      if (t.compareTo(ebt)==0)
        throw new TIRExceptCreate("This tag exists");
      et = idl.tag();
      for (i=0;i<et.length;i++) {
        if (t.compareTo(et[i])==0)
          throw new TIRExceptCreate("This tag exists");       
      } 
    }
    ((SIdentification) id).tag(t);
  }

  private void fromNormObj() throws RemoteException, TIRExceptLock {
    if (normal==null)
      return;
    if (((WorkRepositoryImpl)rep).lock.isLocked())
      throw new TIRExceptLock("Repository is locked");
    int i;
    LiItem akt;
    akt = normal.firstChild;
    for(i=0;i<normal.numOfItems;i++) {
      addListItem(WorkRepositoryImpl.newWorkFromNormal(akt.obj,rep,this,workId));
      akt = akt.next;
    }
    hasCont = true;
  }

  public UnionMember[] members() throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
    } else {
      if (!hasCont) 
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
    }
    UnionMember[] all = new UnionMember [numOfItems];
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (akt.obj.get_def_kind().value()==DefinitionKind.dk_UnionMember) {
        all[j] = (UnionMember) akt.obj;
	j++;
      }
      akt = akt.next;
    }
    UnionMember[] ret = new UnionMember [j];
    for(i=0;i<j;i++) {
      ret[i] = all[i];
    }
    return ret;   
  }

  public Contained[] contents(DefinitionKind type) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.contents(type);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.contents(type);     
    }
  }    

  public Contained lookup(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(id);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(id);
    }
  }

  public Contained lookup(String name, String ver) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(name, ver);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(name, ver);
    }
  }

  public Contained[] lookup_name(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(id);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(id);
    }
  }

  public Contained[] lookup_name(String name) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(name);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(name);
    }
  }

  public Contained lookup_tag(String name, String tag) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_tag(name, tag);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_tag(name, tag);
    }
  }

  public Contained[] lookup_branchtag(String name, String brtag) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_branchtag(name, brtag);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_branchtag(name, brtag);
    }
  }

  public Contained lookup_lastinbranch(String name, String brtag) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastinbranch(name, brtag);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastinbranch(name, brtag);
    }
  }
  
  public Contained lookup_lastfromversion(String name, String version) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastfromversion(name, version);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastfromversion(name, version);
    }
  }

  public Contained[] scontents(DefinitionKind type) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      return super.scontents(type);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.scontents(type);
    }
  }

  public Contained slookup(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
        return super.slookup(id);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup(id);
    }
  }

  public Contained slookup(String name, String version) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
        return super.slookup(name, version);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup(name, version);
    }
  }

  public Contained[] slookup_name(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
        return super.slookup_name(id);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup_name(id);
    }
  }

  public Contained[] slookup_name(String name) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      return super.slookup_name(name);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup_name(name);
    }
  }

  public UnionMember create_member(String name, CDLType type, ExprOperDef[] label) throws RemoteException, TIRExceptCreate, TIRExceptLock {
  /* !!! dopsat testovani typu u label */
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (name!=null && name.compareTo("")!=0 && type!=null && label!=null && label.length!=0) {
      if (type.get_state().value()==StateKind.sk_work) {
        for (int i=0;i<label.length;i++) {
          if (label[i].get_state().value()!=StateKind.sk_work) {
            throw new TIRExceptCreate("Given object isn't work.");
          }
        }  
	UnionMember[] a = members();
	boolean found = false;
	for (int i=0;i<a.length;i++) {
          if (a[i].name().compareTo(name)==0) {
	    found = true;
	    break;
	  }
	}
        if (found)
	  throw new TIRExceptCreate("Member with this name exists in union.");
        ExprOperDef[] ll = new ExprOperDef [label.length];
        for (int i=0;i<label.length;i++) 
          ll[i] = label[i];
	UnionMember m = new UnionMemberImpl(new IdentificationImpl(((SIdentification) this.id).language(), this.id.absolute_name().name()+"::"+name,this.id.version()), this, rep, type, ll);
	addListItem(m);
	return m;
      } else {
        throw new TIRExceptCreate("Given object isn't work.");
      }
    } else {
      return null;
    }
  }

  public StructDef create_struct(Identification id) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id!=null) {
      if (!hasCont)
        fromNormObj();
      // set version of substructure same as this version  
      id = new IdentificationImpl(((SIdentification) id).language(), id.absolute_name().name(),this.id.version());
      Contained[] a = lookup_name(id.name());
      if (a.length != 0) {
        if (a[0].get_def_kind().value() != DefinitionKind.dk_Struct)
	  throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	boolean found = false;  
	for (int i=0;i<a.length;i++) {
	  if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	    found = true;
	    break;
	  }
	}
	if (found) 
	  throw new TIRExceptCreate("Object with same identification exists in repository.");
      }
      // creating new object
      StructDef ret = new StructDefImpl(id, this, rep, workId);
      addListItem(ret); // adding object to list
      return ret;       // return object
    } else
      return null;
  }
  
  public UnionDef create_union(Identification id, CDLType switch_type) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id==null || switch_type==null)
      return null;
    else {
      if (switch_type.get_state().value()==StateKind.sk_work ) { 
        if (!hasCont)
          fromNormObj();
        // set version of substructure same as this version  
        id = new IdentificationImpl(((SIdentification) id).language(), id.absolute_name().name(),this.id.version());
        Contained[] a = lookup_name(id.name());
        if (a.length != 0) {
          if (a[0].get_def_kind().value() != DefinitionKind.dk_Union)
	    throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	  boolean found = false;  
	  for (int i=0;i<a.length;i++) {
	    if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	      found = true;
	      break;
	    }
	  }
	  if (found) 
	    throw new TIRExceptCreate("Object with same identification exists in repository.");
        }
        // creating new object
        UnionDef ret = new UnionDefImpl(id, this, rep, switch_type,workId);
        addListItem(ret); // adding object to list
        return ret;       // return object
      } else {
        throw new TIRExceptCreate("Given object isn't work.");
      }
    }
  }
  
  public EnumDef create_enum(Identification id) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id!=null) {
      if (!hasCont)
        fromNormObj();
      // set version of substructure same as this version  
      id = new IdentificationImpl(((SIdentification) id).language(), id.absolute_name().name(),this.id.version());
      Contained[] a = lookup_name(id.name());
      if (a.length != 0) {
        if (a[0].get_def_kind().value() != DefinitionKind.dk_Enum)
	  throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	boolean found = false;  
	for (int i=0;i<a.length;i++) {
	  if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	    found = true;
	    break;
	  }
	}
	if (found) 
	  throw new TIRExceptCreate("Object with same identification exists in repository.");
      }
      // creating new object
      EnumDef ret = new EnumDefImpl(id, this, rep, workId);
      addListItem(ret); // adding object to list
      return ret;       // return object
    } else
      return null;
  }

  /*from interface TIRImplObject */
  public void save(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      boolean closeFile = false;
      if (st.curOutFile==null) {  // saving in module
        st.curOutFile = new DataOutputStream(new FileOutputStream(st.currentFile));
	closeFile = true;
      } else {   // saving in other structure (or union)
        ;
      }
      ((IdentificationImpl)id).save(st.curOutFile);   // saving identification of union
      Storage.writeCDLType(st, swtype);
      st.curOutFile.writeInt(numOfItems);
      LiItem akt = firstChild;
      for(int i=0;i<numOfItems;i++) {  // saving all childs
        st.curOutFile.writeInt(akt.obj.get_def_kind().value());
        ((TIRImplObject)(akt.obj)).save(st);
	akt = akt.next;
      }
      if (closeFile)  { // close file 
        st.curOutFile.close();
	st.curOutFile = null;
      }
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in "+st.current+".");
    }
  }

  /*from interface TIRImplObject */
  public void load(Storage st) throws RemoteException, TIRExceptStorage{
    try {
      boolean closeFile = false;
      if (st.curInFile==null) {
        st.curInFile = new DataInputStream(new FileInputStream(st.currentFile));
        closeFile = true;
      } else {
        ;
      }
      id = new IdentificationImpl();
      ((IdentificationImpl)id).load(st.curInFile);
      swtype = Storage.readCDLType(st);
      int num = st.curInFile.readInt();
      int dkind;
      TIRImplObject obj;
      for (int i=0;i<num;i++) {
        dkind = st.curInFile.readInt();
	switch (dkind) {
	  case DefinitionKind.dk_UnionMember:
	    obj = new UnionMemberImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
          case DefinitionKind.dk_Struct:
	    obj = new StructDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Enum:
	    obj = new EnumDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Union:
	    obj = new UnionDefImpl(this,rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  default:
            throw new TIRExceptStorage("Unexpected kind of object in \""
	                              +st.currentFile+"\".");
	} // -- end of switch
      }
      if (closeFile)  { // close file 
        st.curInFile.close();
	st.curInFile = null;
      }
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in "+st.current+".");
    }
  }

  public void postLoad(RepositoryImpl r) throws RemoteException, TIRExceptStorage {
    try {
    
    if (swtype.get_def_kind().value() == DefinitionKind.dk_none) {   // is NotYetInRep?
      if (((NotYetInRep) swtype).defKind == DefinitionKind.dk_Primitive)
        swtype = r.get_prim(((NotYetInRep) swtype).primKind);
      else {
        AbsoluteName nm = ((NotYetInRep) swtype).nm;
	Container con = r;  // begin search from repository
	Contained[] rt;
	Contained ret;
	IdentificationImpl id;
        for (int i=0;i<nm.size()-1;i++) {  // last element is name of object
	  id = new IdentificationImpl(nm.elementAt(0),"::"+nm.elementAt(i),"");
          rt = ((SContainer)con).slookup_name(id);
	  if (rt.length==0) throw new TIRExceptStorage("Can't find type " + nm.name());
          if (rt.length==1) {
            con = (Container) rt[0];
	  } else {  // type in struct (etc.) - more versions of this struct
	    boolean found = false;
            id.version(((NotYetInRep) swtype).id.version());  // substruct version is inherited from parent
            int j = 0;
	    for (j=0;j<rt.length;j++) {
	      if (id.is_short_equal(rt[j].get_identification())) {
	        found = true;
		break;  // -- break for
	      }
	    } // -- end of for
	    if (!found) throw new TIRExceptStorage("Can't find type " + nm.name());
	    con = (Container) rt[j];
	  }
	}
	if ((ret = ((SContainer)con).slookup(((NotYetInRep) swtype).id))==null)
	  throw new TIRExceptStorage("Can't find type " + nm.name());
	swtype = (CDLType) ret;
	if (swtype.get_def_kind().value()==DefinitionKind.dk_Array ||
            swtype.get_def_kind().value()==DefinitionKind.dk_Sequence)
	  ((TIRImplObject) swtype).postLoad(r);
      }
    }
    // postLoad() items
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      ((TIRImplObject) akt.obj).postLoad(r);
      akt = akt.next;
    }

    } catch (TIRExceptLock e) {;}
  }

  public boolean isNew() {
    return ((stKindImpl.value()==StateKind.sk_work) && normal==null);
  }

  public void canCommit() throws RemoteException, TIRExceptCommit {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    LiItem akt;
    int i;
    akt = firstChild;
    for (i=0;i<numOfItems;i++) {
      ((TIRImplObject)akt.obj).canCommit();
      akt = akt.next;
    }
  }

  public void doCommit(Container in, Repository rep) throws RemoteException {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    if (isNew()) {  
      stKindImpl.toNormal();
      switch (swtype.get_def_kind().value()) {
        case DefinitionKind.dk_Primitive:
          swtype = ((SRepository)rep).get_prim(((PrimitiveDef)swtype).kind().value());
          break;
        case DefinitionKind.dk_Array:
          ArrayDef nt = new ArrayDefImpl(ExprHelp.toNormal(((ArrayDef) swtype).length()), ((ArrayDef) swtype).element_type());
          swtype = nt;
          ((TIRImplObject)swtype).doCommit(in,rep);
          break;
        case DefinitionKind.dk_Sequence:
          SequenceDef st = new SequenceDefImpl(ExprHelp.toNormal(((SequenceDef) swtype).bound()), ((SequenceDef) swtype).element_type());
          swtype = st;
          ((TIRImplObject)swtype).doCommit(in,rep);
          break;
        case DefinitionKind.dk_String:
          StringDef str = new StringDefImpl(ExprHelp.toNormal(((StringDef) swtype).bound()));
          swtype = str;
          ((TIRImplObject)swtype).doCommit(in,rep);
          break;
        case DefinitionKind.dk_Wstring:
          WstringDef wstr = new WstringDefImpl(ExprHelp.toNormal(((WstringDef) swtype).bound()));
          swtype = wstr;
          ((TIRImplObject)swtype).doCommit(in,rep);
          break;
        case DefinitionKind.dk_Fixed:
          FixedDef ft = new FixedDefImpl(ExprHelp.toNormal(((FixedDef) swtype).digits()), ExprHelp.toNormal(((FixedDef) swtype).scale()));
          swtype = ft;
          ((TIRImplObject)swtype).doCommit(this,rep);
          break;
        default:
          Identification idl = ((Contained)swtype).get_identification();
          swtype = new NotYetInRep(idl.lang_absolute_name().name(), idl.name(), idl.version());
          break;
      }
      LiItem akt = firstChild;
      for (int i=0;i<numOfItems;i++) {
        ((TIRImplObject)akt.obj).doCommit(this,rep);
        akt = akt.next;     
      }
      parent = in;
      this.rep = rep;
    } else {
      normal = null;
    }
  }

  public void doAbort(long workId) throws RemoteException {
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      if (((TIRImplObject)akt.obj).isNew()) 
        ((TIRImplObject)akt.obj).doAbort(workId);
      akt = akt.next;
    }  
  }
}
