/* $Id: InstallToTR.java,v 1.3 2004/03/22 19:31:49 hnetynka Exp $ */
package SOFA.SOFAnode.Made.Tools;
import java.io.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import org.xml.sax.*;

/** Install component to "TR".
  * Arguments:
  * <ul>
  * <li>1st argument - ddform file (it was created by CodeGen and completed by developer</li>
  * <li>next arguments - class files to pack</li>
  * </ul>
  * Java properties:
  * <ul>
  * <li>file.root - from which directory (inserted) files will be searched
  * <li>sofa.tr.dir - directory with TR
  * </ul>
  *
  * @author Petr Hnetynka
  */
public class InstallToTR {
  
  private static Element dd;
  private static String version;
  private static String arch;
  private static java.util.Hashtable ent;
  private static java.util.Hashtable fullNamesWithoutVersion;
  private static java.util.Properties trIndex;
  private static java.util.Properties implIndex;
  private static String lastIndex;

  public static void main(String[] argv) {
    if (argv.length < 2) {
      System.out.println("too few arguments");
      System.exit(1);
    }
  

    String trDir = System.getProperty("sofa.tr.dir", null);
    if (trDir == null) {
      System.out.println("Specify TR directrory (java property \"sofa.tr.dir\")");
      System.exit(1);
    }
    trIndex = new java.util.Properties();
    implIndex = new java.util.Properties();
    try {
      FileInputStream fis = new FileInputStream(trDir+File.separator+"iface"+File.separator+"index");
      trIndex.load(fis);
      fis.close();

      fis = new FileInputStream(trDir+File.separator+"impl"+File.separator+"info");
      implIndex.load(fis);
      fis.close();

      fis = new FileInputStream(trDir+File.separator+"iface"+File.separator+"last");
      StringBuffer sb = new StringBuffer();
      int a = fis.read();
      while (a != -1) {
        sb.append((char) a);
        a = fis.read();
      }
      fis.close();
      lastIndex = sb.toString();
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    }
    
    String fileRoot = System.getProperty("file.root", ".");
    fileRoot += File.separator;

    try {
      dd = getDDElem(fileRoot+argv[0]);
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      System.out.println("ParserConfigurationException: "+e.getMessage() );
      System.exit(1);
    } catch (org.xml.sax.SAXException e) {
      System.out.println("SAXException: "+e.getMessage() );
      System.exit(1);
    }
    
    version = getImplVersion(dd);
    ent = getCDLEntities(dd);
    arch = getArchitectureFullName(dd);
    fullNamesWithoutVersion = getFullNamesWithoutVersion(trIndex);

    // prepare translation table
    java.util.Hashtable trTable = new java.util.Hashtable();
    // add interfaces to trTable
    for (java.util.Enumeration e = ent.keys(); e.hasMoreElements(); ) {
      String javaName = (String) e.nextElement();
      String fullName = (String) ent.get(javaName);
      String specVersion = fullName.substring(fullName.indexOf('?')+1);
      trTable.put(toVMName(javaName), toVMName(javaName+specVersionToFileName(specVersion)));
    }
    // add implementation classes to trTable
    for (int i=1; i<argv.length; i++) {
      try {
        SOFA.Tools.Resolver.ClassInfo info = new SOFA.Tools.Resolver.ClassInfo(argv[i]);
        trTable.put(info.className(), info.className()+implVersionToFileName(version));
      } catch (java.io.IOException ex) {
        System.out.println("IOException: "+ex.getMessage());
        System.exit(1);
      }
    }
    // create resolver
    SOFA.Tools.Resolver.Resolver resolver = new SOFA.Tools.Resolver.Resolver(trTable);

    System.out.print("Installing to TR....");

    // copy interfaces
    for (java.util.Enumeration e = ent.keys(); e.hasMoreElements(); ) {
      String javaName = (String) e.nextElement();
      String fullName = (String) ent.get(javaName);
      String withoutVersion = fullName.substring(0, fullName.indexOf('?'));
      String specVersion = fullName.substring(fullName.indexOf('?')+1);
      if (fullNamesWithoutVersion.containsKey(withoutVersion)) {
        if (trIndex.containsKey(fullName)) { 
          // contains same version -> rewrite
          String ind = (String) trIndex.get(fullName);
          File fsource = new File(fileRoot+ classNameToFileName(javaName));
          if (!fsource.exists()) {
            System.out.println("Error: file "+fsource.toString()+" does not exist.");
            System.exit(1); 
          }
          File fdest = new File(trDir+File.separator+"iface"+File.separator+ind+File.separator+classNameToFileName(javaName));
          File dir = fdest.getParentFile();
          if (dir != null) 
            if (! dir.exists()) 
              dir.mkdirs();
          try {
            FileInputStream fis = new FileInputStream(fsource);
            FileOutputStream fos = new FileOutputStream(fdest);
            int a;
            while ( (a = fis.read()) != -1)
              fos.write(a);
            fis.close();
            fos.close();
          } catch (java.io.IOException ex) {
            System.out.println("IOException: "+ex.getMessage());
            System.exit(1);
          }
          // save augmented class
          fdest = new File(trDir+File.separator+"classes"+File.separator+classNameToFileName(javaName+specVersionToFileName(specVersion)));
          dir = fdest.getParentFile();
          if (dir != null) 
            if (! dir.exists()) 
              dir.mkdirs();
          try {
            FileInputStream fis = new FileInputStream(fsource);
            byte [] b = resolver.process(fis);
            FileOutputStream fos = new FileOutputStream(fdest);
            fos.write(b);
            fis.close();
            fos.close();
          } catch (java.io.IOException ex) {
            System.out.println("IOException: "+ex.getMessage());
            System.exit(1);
          }
        } else { //another version
          lastIndex = incrementLastIndex(lastIndex);
          File fsource = new File(fileRoot+ classNameToFileName(javaName));
          if (!fsource.exists()) {
            System.out.println("Error: file "+fsource.toString()+" does not exist.");
            System.exit(1); 
          }
          File fdest = new File(trDir+File.separator+"iface"+File.separator+lastIndex+File.separator+classNameToFileName(javaName));
          File dir = fdest.getParentFile();
          if (dir != null) 
            if (! dir.exists()) 
              dir.mkdirs();
          try {
            FileInputStream fis = new FileInputStream(fsource);
            FileOutputStream fos = new FileOutputStream(fdest);
            int a;
            while ( (a = fis.read()) != -1)
              fos.write(a);
            fis.close();
            fos.close();
          } catch (java.io.IOException ex) {
            System.out.println("IOException: "+ex.getMessage());
            System.exit(1);
          }
          // save augmented class
          fdest = new File(trDir+File.separator+"classes"+File.separator+classNameToFileName(javaName+specVersionToFileName(specVersion)));
          dir = fdest.getParentFile();
          if (dir != null) 
            if (! dir.exists()) 
              dir.mkdirs();
          try {
            FileInputStream fis = new FileInputStream(fsource);
            byte [] b = resolver.process(fis);
            FileOutputStream fos = new FileOutputStream(fdest);
            fos.write(b);
            fis.close();
            fos.close();
          } catch (java.io.IOException ex) {
            System.out.println("IOException: "+ex.getMessage());
            System.exit(1);
          }
          trIndex.put(fullName, lastIndex);
        }
      } else {
        File fsource = new File(fileRoot+ classNameToFileName(javaName));
        if (!fsource.exists()) {
          System.out.println("Error: file "+fsource.toString()+" does not exist.");
          System.exit(1); 
        }
        File fdest = new File(trDir+File.separator+"iface"+File.separator+"a"+File.separator+classNameToFileName(javaName));
        File dir = fdest.getParentFile();
        if (dir != null) 
          if (! dir.exists()) 
            dir.mkdirs();
        try {
          FileInputStream fis = new FileInputStream(fsource);
          FileOutputStream fos = new FileOutputStream(fdest);
          int a;
          while ( (a = fis.read()) != -1)
            fos.write(a);
          fis.close();
          fos.close();
        } catch (java.io.IOException ex) {
          System.out.println("IOException: "+ex.getMessage());
          System.exit(1);
        }
        // save augmented class
        fdest = new File(trDir+File.separator+"classes"+File.separator+classNameToFileName(javaName+specVersionToFileName(specVersion)));
        dir = fdest.getParentFile();
        if (dir != null) 
          if (! dir.exists()) 
            dir.mkdirs();
        try {
          FileInputStream fis = new FileInputStream(fsource);
          byte [] b = resolver.process(fis);
          FileOutputStream fos = new FileOutputStream(fdest);
          fos.write(b);
          fis.close();
          fos.close();
        } catch (java.io.IOException ex) {
          System.out.println("IOException: "+ex.getMessage());
          System.exit(1);
        }
        trIndex.put(fullName, "a");
      }
    }

    // copy implementation files
    String sdir = fullNameToFileName(arch)+File.separator+version+File.separator;
    for (int i=1; i<argv.length; i++) {
      File fsource = new File(fileRoot+ File.separator+argv[i]);
      if (!fsource.exists()) {
        System.out.println("Error: file "+fsource.toString()+" does not exist.");
        System.exit(1); 
      }
      File fdest = new File(trDir+File.separator+"impl"+File.separator+sdir+argv[i]);
      File dir = fdest.getParentFile();
      if (dir != null) 
        if (! dir.exists()) 
          dir.mkdirs();
      try {
        FileInputStream fis = new FileInputStream(fsource);
        FileOutputStream fos = new FileOutputStream(fdest);
        int a;
        while ( (a = fis.read()) != -1)
          fos.write(a);
        fis.close();
        fos.close();
      } catch (java.io.IOException ex) {
        System.out.println("IOException: "+ex.getMessage());
        System.exit(1);
      }
      // save augmented class
      fdest = new File(trDir+File.separator+"classes"+File.separator+argv[i].substring(0, argv[i].length()-6)+implVersionToFileName(version)+".class");
      dir = fdest.getParentFile();
      if (dir != null) 
        if (! dir.exists()) 
          dir.mkdirs();
      try {
        FileInputStream fis = new FileInputStream(fsource);
        byte [] b = resolver.process(fis);
        FileOutputStream fos = new FileOutputStream(fdest);
        fos.write(b);
        fis.close();
        fos.close();
      } catch (java.io.IOException ex) {
        System.out.println("IOException: "+ex.getMessage());
        System.exit(1);
      }
    }
    // copy dd
    File fsource = new File(fileRoot+ File.separator+argv[0]);
    if (!fsource.exists()) {
      System.out.println("Error: file "+fsource.toString()+" does not exist.");
      System.exit(1); 
    }
    File fdest = new File(trDir+File.separator+"impl"+File.separator+sdir+"index.dc");
    try {
      FileInputStream fis = new FileInputStream(fsource);
      FileOutputStream fos = new FileOutputStream(fdest);
      int a;
      while ( (a = fis.read()) != -1)
        fos.write(a);
      fis.close();
      fos.close();
    } catch (java.io.IOException ex) {
      System.out.println("IOException: "+ex.getMessage());
      System.exit(1);
    }
    implIndex.put(arch+"["+version+"]", "1");

    // save indexes
    try {
      FileOutputStream fos = new FileOutputStream(trDir+File.separator+"iface"+File.separator+"last");
      for (int i=0; i<lastIndex.length(); i++) {
        fos.write((byte)lastIndex.charAt(i));
      }
      fos.close();
      fos = new FileOutputStream(trDir+File.separator+"iface"+File.separator+"index");
      trIndex.store(fos, null);
      fos.close();
      fos = new FileOutputStream(trDir+File.separator+"impl"+File.separator+"info");
      implIndex.store(fos, null);
      fos.close();
    } catch (java.io.IOException ex) {
      System.out.println("IOException: "+ex.getMessage());
      System.exit(1);
    }
    
    System.out.println("OK");
  }

  /** Return document element of DD.
    * 
    * @param fname filename
    * @return  document element
    */
  static Element getDDElem(String fname) throws java.io.IOException, ParserConfigurationException, SAXException {
    java.io.File file = new java.io.File(fname);
    if (!file.exists()) {
      throw new java.io.IOException("File \""+fname+"\" doesn't exist.");
    }

    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document document = builder.parse(file);
    Element element = document.getDocumentElement();
    element.normalize();
    return element;
  }

  static String getImplVersion(Element element) {
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType() == Node.ELEMENT_NODE) && (n.getNodeName().compareTo("version")==0)) {
        return n.getFirstChild().getNodeValue();
      }
    }
    return null;
  }

  static String getArchitectureFullName(Element element) {
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType() == Node.ELEMENT_NODE) && (n.getNodeName().compareTo("architecture_ref")==0)) {
        String ar = n.getFirstChild().getNodeValue();
        ar = ar.substring(0,ar.indexOf('?'));
        return ar;
      }
    }
    return null;
  }

  static java.util.Hashtable getCDLEntities(Element element) {
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType() == Node.ELEMENT_NODE) && (n.getNodeName().compareTo("cdl_entities")==0)) {
        NodeList nl1 = n.getChildNodes();
        java.util.Hashtable ret = new java.util.Hashtable();
        for (int j=0; j<nl1.getLength(); j++) {
          Node n1 = nl1.item(j);
          if ((n1.getNodeType() == Node.ELEMENT_NODE) && (n1.getNodeName().compareTo("entity")==0)) {
            ret.put(((Element)n1).getAttribute("javaname"), ((Element)n1).getAttribute("cdlname"));
          }
        }
        return ret;
      }
    }
    return null;
  }

  private static java.util.Hashtable getFullNamesWithoutVersion(java.util.Properties prop) {
    java.util.Hashtable ret = new java.util.Hashtable();
    for (java.util.Enumeration e = prop.keys() ; e.hasMoreElements() ;) {
      String k = (String) e.nextElement();
      k = k.substring(0, k.indexOf('?'));
      ret.put(k, "");
    }
    return ret;
  }

  private static boolean compareFullNameWithoutVersion(String n1, String n2) {
    n1 = n1.substring(0, n1.indexOf('?'));
    n2 = n2.substring(0, n2.indexOf('?'));
    if (n1.compareTo(n2) == 0)
      return true;
    return false;
  }

  /** From name of class creates name of file (replaces dot to slash).
    */
  static String classNameToFileName(String clname) {
    return clname.replace('.', File.separatorChar) + ".class";
  }

  /** From full name creates name of file (replaces doble colon to slash).
    */
  static String fullNameToFileName(String clname) {
    StringBuffer sb = new StringBuffer();
    boolean s = true;
    for (int i=0; i<clname.length(); i++) {
      if (clname.charAt(i)==':') {
        if (s) {
          sb.append(":");
          s = false;
        }
      } else {
        s = true;
        sb.append(clname.charAt(i));
      }
    }
//    System.out.println(sb);
//    return clname.replaceAll("::", String.valueOf(File.separatorChar));
    return sb.toString().replace(':', File.separatorChar);
  }

  private static String incrementLastIndex(String last) {
    StringBuffer sb = new StringBuffer(last);
    int now = last.length() - 1;
    while (true) {
      if (last.charAt(now) != 'z') {
        sb.setCharAt(now, (char) ((char)last.charAt(now)+(char)1));
        return sb.toString();
      }
      sb.setCharAt(now, 'a');
      now--;
      if (now < 0) {
        sb.append('a');
        return sb.toString();
      }
    }
  }

  static String specVersionToFileName(String specVersion) {
    StringBuffer ret = new StringBuffer("_Q_");
    for (int i=0; i<specVersion.length(); i++) {
      switch (specVersion.charAt(i)) {
      case '.': 
        ret.append('_');
        break;
      case '!':
        ret.append("_E_");
        break;
      default:
        ret.append(specVersion.charAt(i));
      }
    }
    return ret.toString();
  }

  static String toVMName(String name) {
    StringBuffer ret = new StringBuffer();
    for (int i=0; i<name.length(); i++) {
      if (name.charAt(i) == '.')
        ret.append('/');
      else
        ret.append(name.charAt(i));
    }
    return ret.toString();
  }

  static String implVersionToFileName(String implVersion) {
    StringBuffer ret = new StringBuffer("_V_");
    for (int i=0; i<implVersion.length(); i++) {
      switch (implVersion.charAt(i)) {
      case '.': 
        ret.append('_');
        break;
      case '!':
        ret.append("_E_");
        break;
      default:
        ret.append(implVersion.charAt(i));
      }
    }
    return ret.toString();
  }
}
