/* $Id: MakeAppl.java,v 1.2 2003/03/17 10:56:19 hnetynka Exp $ */
package SOFA.SOFAnode.Made.Tools;
import java.io.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import SOFA.SOFAnode.Run.SpecFile;
import SOFA.SOFAnode.Made.TIR.*;
import SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods;

/** Make application from separated depldescriptors in TR.
  * 1st argument is full application name 
  *
  * @author Petr Hnetynka
  */
public class MakeAppl {
  private static String trImplRoot;
  
  public static void main(String[] argv) {
    if (argv.length < 1) {
      System.out.println("Too few arguments.");
      System.out.println("  1st argument - name of the specfile");
      System.exit(1);
    }
    
    SpecFile specFile = null;
    try {
      specFile = SpecFile.getSpecFileFromFile(argv[0]);
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      System.out.println("ParserConfigurationException: "+e.getMessage() );
      System.exit(1);
    } catch (org.xml.sax.SAXException e) {
      System.out.println("SAXException: "+e.getMessage() );
      System.exit(1);
    }

    Element specFileElement = specFile.getRootElement();
    
    String trDir = System.getProperty("sofa.tr.dir", null);
    if (trDir == null) {
      System.out.println("Specify TR directrory (java property \"sofa.tr.dir\")");
      System.exit(1);
    }

    trImplRoot = trDir + File.separator + "impl" + File.separator;
    

    Repository tir = null;
    try {
      tir = TIRAccessMethods.getRepository();
    } catch (java.net.MalformedURLException e) {
      System.out.println("MalformedURLException: "+e.getMessage() );
      System.exit(1);
    } catch (java.rmi.NotBoundException e) {
      System.out.println("NotBoundException: "+e.getMessage() );
      System.exit(1);
    } catch (java.rmi.RemoteException e) {
      System.out.println("RemoteException: "+e.getMessage() );
      System.exit(1);
    }
    
    processSpecFileElement(tir, null, specFileElement);

    try {
      TransformerFactory tFactory = TransformerFactory.newInstance();
      Transformer transformer = tFactory.newTransformer();
      transformer.setOutputProperty("indent", "yes");
      transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "5");
      DOMSource source = new DOMSource(specFileElement);
      FileOutputStream out = new FileOutputStream( trImplRoot + fullNameToFileName(SpecFile.getFullApplName(specFileElement), "run.dc"));
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
    } catch  (TransformerConfigurationException e) {
      System.out.println("TransformerConfigurationException: "+e.getMessage() );
      System.exit(1);
    } catch (TransformerException e) {
      System.out.println("TransformerException: "+e.getMessage() );
      System.exit(1);
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    }
    
  }

  private static void processSpecFileElement(Repository tir, SOFA.SOFAnode.Run.GenSpecFile.InfoArray riarray, Element specFileElem) {
    Document specDoc = specFileElem.getOwnerDocument();
    String fullName = SpecFile.getFullApplName(specFileElem);
//    System.out.println(fullName);
    String archName = SpecFile.getFullArchName(specFileElem);
    String inst = specFileElem.getAttribute("name");
    SOFA.SOFAnode.Run.GenSpecFile.InfoArray iarray = SOFA.SOFAnode.Run.GenSpecFile.getBindConnInfo(tir, archName);
    Element archRootElement = null;
    try {
      archRootElement = InstallToTR.getDDElem(trImplRoot + fullNameToFileName(fullName, "index.dc") );
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      System.out.println("ParserConfigurationException: "+e.getMessage() );
      System.exit(1);
    } catch (org.xml.sax.SAXException e) {
      System.out.println("SAXException: "+e.getMessage() );
      System.exit(1);
    }
    java.util.ArrayList cons = getAllElements(specFileElem, "connector");
    for (int i=0; i<cons.size(); i++) {
      processCon(riarray, inst, (Element) cons.get(i));
      removeTextNodes((Element) cons.get(i));
    }
    String version = getElement(specFileElem, "version").getFirstChild().getNodeValue();
    Node specFileNextVerNode = getElement(specFileElem, "version").getNextSibling();
    Element cdlEnts = getElement(archRootElement, "cdl_entities");
    removeTextNodes(cdlEnts);
    java.util.ArrayList impls = getAllElements(archRootElement, "implementation");
    specFileElem.insertBefore(specDoc.importNode(cdlEnts, true), specFileNextVerNode);
    for (int i=0; i<impls.size(); i++) {
      Element implEl = (Element) specDoc.importNode((Element) impls.get(i), true);
     /* if (implEl.getAttribute("type").compareTo("builder")==0) {
        Node fn = implEl.getFirstChild();
        if (fn != null) {
          String b = fn.getNodeValue();
          b += InstallToTR.implVersionToFileName(version);
          fn.setNodeValue(b);
        }
      } */
      specFileElem.insertBefore(implEl, specFileNextVerNode);
    }
    java.util.ArrayList comps = getAllElements(specFileElem, "sofa_component");
    for (int i=0; i<comps.size(); i++) {
      processSpecFileElement(tir, iarray, (Element) comps.get(i));
    }
    java.util.ArrayList units = getAllElements(specFileElem, "unit");
    for (int i=0; i<units.size(); i++) {
      comps = getAllElements((Element) units.get(i) , "sofa_component");
      for (int j=0; j<comps.size(); j++) {
        processSpecFileElement(tir, iarray, (Element) comps.get(j));
      }
    }
  }

  private static Element addApplName(Element element) {
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo("architecture_ref")==0)) {
        String ar = n.getFirstChild().getNodeValue();
        ar = ar.substring(ar.lastIndexOf(':')+1, ar.indexOf('?'));
        element.setAttribute("name", ar);
      }
    }
    return element;
  }

  private static String fullNameToFileName(String clname, String fname) {
//    clname = clname.replaceAll("::", File.separator);
    StringBuffer sb = new StringBuffer();
    boolean s = true;
    for (int i=0; i<clname.length(); i++) {
      if (clname.charAt(i)==':') {
        if (s) {
          sb.append(":");
          s = false;
        }
      } else {
        s = true;
        sb.append(clname.charAt(i));
      }
    }
    clname = sb.toString().replace(':', File.separatorChar);
    clname = clname.replace('[', File.separatorChar);
    clname = clname.replace(']', File.separatorChar);
    clname = clname + fname;
    return clname;
  }
  
  private static Element getElement(Element element, String name) {
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo(name)==0)) {
        return (Element) n;
      }
    }
    return null;
  }
  
  private static java.util.ArrayList getAllElements(Element element, String name) {
    java.util.ArrayList ret = new java.util.ArrayList();
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo(name)==0)) {
        ret.add(n);
      }
    }
    return ret;
  }

  private static final int CSR = 0;
  private static final int CSP = 1;
  private static final int EPR = 2;
  private static final int EPP = 3;
  private static final int DSR = 4;
  private static final int DSP = 5;

  private static void processCon(SOFA.SOFAnode.Run.GenSpecFile.InfoArray riarray, String inst, Element conEl) {
    Document doc = conEl.getOwnerDocument();
    String connName = conEl.getAttribute("name");
    String connFrame = conEl.getAttribute("frame");
//        System.out.println(connFrame);

    int ckind = 0;
    if (connName.startsWith("_delegate_")) {
      conEl.removeAttribute("frame");
      conEl.setAttribute("impl", "SGenerator:CSProcCallImpl");
      conEl.setAttribute("unit", "Client");
      ckind = CSR;
    } else if (connName.startsWith("_subsume_")) {
      conEl.removeAttribute("frame");
      conEl.setAttribute("impl", "SGenerator:CSProcCallImpl");
      conEl.setAttribute("unit", "Server");
      ckind = CSP;
    } else if (connFrame.compareTo("SGenerator:CSProcCallImpl")==0) {
      conEl.removeAttribute("frame");
      conEl.setAttribute("impl", "SGenerator:CSProcCallImpl");
      SOFA.SOFAnode.Run.GenSpecFile.BindConnInfo bi = riarray.getInfo(inst, connName);
      if (bi == null) {
        System.out.println("Mismatch in TR and TIR.\nCannot find connector for "+inst+":"+connName+".");
        System.exit(1);
      }
      if (bi.type == SOFA.SOFAnode.Run.GenSpecFile.BindConnInfo.PROV) {
        conEl.setAttribute("unit", "Server");
        ckind = CSP;
      } else {
        conEl.setAttribute("unit", "Client");
        ckind = CSR;
      }
    } else if (connFrame.compareTo("SGenerator:EventPassingImpl")==0) {
      conEl.removeAttribute("frame");
      conEl.setAttribute("impl", "SGenerator:EventPassingImpl");
      SOFA.SOFAnode.Run.GenSpecFile.BindConnInfo bi = riarray.getInfo(inst, connName);
      if (bi == null) {
        System.out.println("Mismatch in TR and TIR.\nCannot find connector for "+inst+":"+connName+".");
        System.exit(1);
      }
      if (bi.type == SOFA.SOFAnode.Run.GenSpecFile.BindConnInfo.PROV) {
        conEl.setAttribute("unit", "Consumer");
        ckind = EPP;
      } else {
        conEl.setAttribute("unit", "Supplier");
        ckind = EPR;
      }
    } else if (connFrame.compareTo("SGenerator:DataStreamImpl")==0) {
      conEl.removeAttribute("frame");
      conEl.setAttribute("impl", "SGenerator:DataStreamImpl");
      SOFA.SOFAnode.Run.GenSpecFile.BindConnInfo bi = riarray.getInfo(inst, connName);
      if (bi == null) {
        System.out.println("Mismatch in TR and TIR.\nCannot find connector for "+inst+":"+connName+".");
        System.exit(1);
      }
      if (bi.type == SOFA.SOFAnode.Run.GenSpecFile.BindConnInfo.PROV) {
        conEl.setAttribute("unit", "Receiver");
        ckind = DSP;
        conEl.setAttribute("requires","::SOFA::Connector::EEG::Types::DataStream::OutputStream?nenya.ms.mff.cuni.cz!0");
        conEl.setAttribute("provides","::SOFA::Connector::EEG::Types::DataStream::InputStream?nenya.ms.mff.cuni.cz!0");
      } else {
        conEl.setAttribute("unit", "Sender");
        ckind = DSR;
        conEl.setAttribute("provides","::SOFA::Connector::EEG::Types::DataStream::OutputStream?nenya.ms.mff.cuni.cz!0");
        conEl.setAttribute("requires","::SOFA::Connector::EEG::Types::DataStream::InputStream?nenya.ms.mff.cuni.cz!0");
      }

    } else {
      System.out.println("Unknown connector frame: "+connFrame);
      System.exit(1);
    }
    // now process transport types
    Element newEl = null;
    NodeList nl = conEl.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if (n.getNodeType()==Node.ELEMENT_NODE) {
        if (n.getNodeName().compareTo("transport")==0) {
          String type = ((Element)n).getAttribute("type");
          switch (ckind) {
          case CSP:
            conEl.removeChild(n);
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "sRole", "impl", "EEM:SRoleImpl"));
            if (type.compareTo("LOCAL")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "skeleton", "impl", "EEM:LocalSkeleton"));
            } else if (type.compareTo("RMI")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "skeleton", "impl", "EEM:RMISkeleton"));
            } else if (type.compareTo("CORBA")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "skeleton", "impl", "EEM:CORBASkeleton"));
            } else {
              System.out.println("Unknown transport type" + type);
              System.exit(1);
            }
            break;
          case CSR:
            conEl.removeChild(n);
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "cRole", "impl", "EEM:CRoleImpl"));
            if (type.compareTo("LOCAL")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "stub", "impl", "EEM:LocalStub"));
            } else if (type.compareTo("RMI")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "stub", "impl", "EEM:RMIStub"));
            } else if (type.compareTo("CORBA")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "stub", "impl", "EEM:CORBAStub"));
            } else {
              System.out.println("Unknown transport type" + type);
              System.exit(1);
            }
            break;
          case EPR:
            conEl.removeChild(n);
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "sRole", "impl", "EEM:CRoleImpl"));
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "distributor", "impl", "EEM:DistributorImpl"));
            if (type.compareTo("LOCAL")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "stub", "impl", "EEM:LocalStub"));
            } else if (type.compareTo("RMI")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "stub", "impl", "EEM:RMIStub"));
            } else if (type.compareTo("CORBA")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "stub", "impl", "EEM:CORBAStub"));
            } else {
              System.out.println("Unknown transport type" + type);
              System.exit(1);
            }
            break;
          case EPP:
            conEl.removeChild(n);
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "cRole", "impl", "EEM:SRoleImpl"));
            if (type.compareTo("LOCAL")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "skeleton", "impl", "EEM:LocalSkeleton"));
            } else if (type.compareTo("RMI")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "skeleton", "impl", "EEM:RMISkeleton"));
            } else if (type.compareTo("CORBA")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "skeleton", "impl", "EEM:CORBASkeleton"));
            } else {
              System.out.println("Unknown transport type" + type);
              System.exit(1);
            }
            break;
          case DSP:
            conEl.removeChild(n);
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "pushRRole", "impl", "EEM:SRoleImpl"));
            if (type.compareTo("LOCAL")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "receiver", "impl", "EEM:DataStreamLocalReceiver"));
              conEl.appendChild(newElemWithTwoAttribs(doc, "property", "name", "senderMode", "value", "push"));
              conEl.appendChild(newElemWithTwoAttribs(doc, "property", "name", "receiverMode", "value", "push"));
            } else {
              System.out.println("DataStream can be only LOCAL.");
              System.exit(1);
            }
            break;
          case DSR:
            conEl.removeChild(n);
            conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "pushSRole", "impl", "EEM:CRoleImpl"));
            if (type.compareTo("LOCAL")==0) {
              conEl.appendChild(newElemWithTwoAttribs(doc, "element", "name", "sender", "impl", "EEM:DataStreamLocalSender"));
              conEl.appendChild(newElemWithTwoAttribs(doc, "property", "name", "senderMode", "value", "push"));
              conEl.appendChild(newElemWithTwoAttribs(doc, "property", "name", "receiverMode", "value", "push"));
            } else {
              System.out.println("DataStream can be only LOCAL.");
              System.exit(1);
            }
            break;
          }
        }
      }
    }
    
  }

  private static Element newElemWithTwoAttribs(Document doc, String elName, String name1, String value1, String name2, String value2) {
    Element newEl =  doc.createElement(elName);
    newEl.setAttribute(name1, value1);
    newEl.setAttribute(name2, value2);
    return newEl;
  }

  private static void removeTextNodes(Element el) {
    NodeList nl = el.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if (n.getNodeType() == Node.TEXT_NODE) {
        el.removeChild(n);
      }
    }
  }
}
