/* $Id: DeplDockImpl.java,v 1.3 2004/03/22 19:31:49 hnetynka Exp $ */
package SOFA.SOFAnode.Run.Deployment;

import SOFA.SOFAnode.Run.SOFARuntimeException;
//import SOFA.SOFAnode.Run.SOFABaseClassLoader;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.InetAddress;
import java.net.UnknownHostException;

import SOFA.SOFAnode.Run.DeplDockRegistry.*;

import org.omg.CosNaming.*;


/** Implementation of deployment descriptor.
  *
  * @author Petr Hnetynka
  */
public class DeplDockImpl implements DeplDock {

  /// Technology description of deployment dock.
  protected TechnologyDescriptor[] description;

  /** Name of the deployment dock. */
  protected String name;
  /** Name of the sofa node */
  protected String node;
  /** Deployment dock registry */
  protected DeplDockRegistry ddr;
  
  /** All CM of components in this dock. */
  protected java.util.Hashtable components;

  /** SRoles of CM in this dock. */
  protected java.util.Hashtable sroles;

  /** All applications (system) in this dock. */
  protected java.util.Hashtable applications;

  /** Constructor.
    * @param _name   name of the deployment dock
    * @param _node   name of the sofanode
    * @param _ddr    deployment dock registry
    */
  public DeplDockImpl(String _name, String _node, DeplDockRegistry _ddr) {
    name = _name;
    node = _node;
    ddr = _ddr;
    components = new java.util.Hashtable();
    sroles = new java.util.Hashtable();
    applications = new java.util.Hashtable();
    defineDescription();
  }

  /** Define the technology <tt>description</tt> of the dock.
    * It is called from constructor and
    * in subclasses, this method should be rewritten to change description.
    * !!! This is only testing implementation !!!
    */
  protected void defineDescription() {
    description = new TechnologyDescriptor [3];
/*    description[0] = new TechnologyProperty("os","any");
    description[1] = new TechnologyProperty("exe","java");
    description[2] = new TechnologyProperty("","");*/
  }

/*  public DeplDockRegistry getDeplDockRegistry() {
    // zrusit !!!
    return null;
  } */
  
  public TechnologyDescriptor[] describeUnderlyingEnvironment() {
    return description;
  }
  
  public SOFA.Connector.Reference instantiate(DeploymentDescriptor dd, String name) throws DeploymentException {
    String id = ddr.getUniqueID();
    ((DeploymentDescriptorImpl) dd).setInstanceId(id);
    if (name == null || name.compareTo("")==0)
      name = dd.getInstanceName();
    String cid = name+":"+id;
    SOFA.Component.DCUP.DCUPComponentManager cm;
    try {
      cm = new SOFA.Component.DCUP.DCUPComponentManagerImpl(dd, node, name, this, id);
    } catch (SOFA.Component.ComponentLifecycleException e) {
      throw new DeploymentException("ComponentLifecycleException: "+e.getMessage(), e);
    }
    SOFA.Connector.RoleBase srole;
/*    try {
      srole = SOFA.Connector.Boot.DCUPComponentManagerConnector.createSrv(cm, id);
    } catch (SOFA.Connector.ConnectorException e) {
      throw new DeploymentException("ConnectorException: "+e.getMessage(), e);
    }*/
    srole = ((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).connToThisCM();
    
    components.put(cid, cm);
    sroles.put(cid, srole);
    if (dd.isSystem())
      applications.put(cid, cm);

    //((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).connToThisCM(srole);
    
    return srole.getSOFAReference();
/*    return new SOFA.Connector.Reference("TEST", new SOFA.Connector.TaggedProfile[] {  new SOFA.Connector.ECG.Profiles.RMIProfile("rmi://localhost/OK") } );  */
  }


  public String[] namesOfAllRunningComponents() {
    String[] ret = new String [components.size()];
    int i = 0;
    for (java.util.Enumeration e = components.keys() ; e.hasMoreElements() ; i++) {
      ret[i] = new String ((String) e.nextElement());
    }
    return ret;
  }
  
  public String[] namesOfAllRunningApplications() {
    String[] ret = new String [applications.size()];
    int i = 0;
    for (java.util.Enumeration e = applications.keys() ; e.hasMoreElements() ; i++) {
      ret[i] = new String ((String) e.nextElement());
    }
    return ret;
  }



  /** Removes LOCAL profile from SOFA reference.
    * @param ref  sofa reference
    * @return     sofa reference without LOCAL profile
    */
  public static SOFA.Connector.Reference removeLocalProfile(SOFA.Connector.Reference ref) {
    boolean b = false;
    for(int i=0; i<ref.profiles.length; i++) {
      if (ref.profiles[i] instanceof SOFA.Connector.ECG.Profiles.LocalProfile) {
        b = true;
      }
    }
    if (b) {
      SOFA.Connector.TaggedProfile[] p = new SOFA.Connector.TaggedProfile [ref.profiles.length];
      int j = 0;
      for (int i=0; i<ref.profiles.length; i++) {
        if (!(ref.profiles[i] instanceof SOFA.Connector.ECG.Profiles.LocalProfile)) {
          p[j] = ref.profiles[i];
          j++;
        }
      }
      ref.profiles = p;
    } 
    return ref;
  }

  /** Converts sofa reference to serialized form.
    * @param ref   sofa reference
    * @return  serialized form of sofa reference or <tt>null</tt>
    */
  public static byte[] sofaReferenceToBytes(SOFA.Connector.Reference ref) {
    try {
//      System.out.println("1");
      cz.cuni.sofa.lib.Impl.ByteArrayOutputStream os = new cz.cuni.sofa.lib.Impl.ByteArrayOutputStream();
//      System.out.println(os);
      ref._write(os);
      return os.toByteArray();
    } catch (java.io.IOException e) {
      return null;
    }
  }

  synchronized public void removeComponent(String name) throws SOFA.SOFAnode.Run.Deployment.DeploymentException {
    if (!sroles.containsKey(name)) 
      throw new DeploymentException("No component with given name");
    SOFA.Component.DCUP.DCUPComponentManagerImpl cm = (SOFA.Component.DCUP.DCUPComponentManagerImpl) components.get(name);
    if (cm == null) {
      sroles.remove(name);
      throw new DeploymentException("Error in the component registering");
    }
    if (!cm.isStopped())
      throw new DeploymentException("Component can't be removed");
    sroles.remove(name);
    components.remove(name);
    if (applications.containsKey(name))
      applications.remove(name);
  }
  
  /** Converts serialized form of sofa reference to real sofa reference.
    * @param ref  byte array with sofa reference
    * @return  sofa reference or <tt>null</tt>
    */
  public static SOFA.Connector.Reference bytesToSofaReference(byte[] ref) {
    try {
      cz.cuni.sofa.lib.Impl.ByteArrayInputStream is = new cz.cuni.sofa.lib.Impl.ByteArrayInputStream (ref);
      SOFA.Connector.Reference ret = new SOFA.Connector.Reference ();
      ret._read(is);
      return ret;
    } catch (java.io.IOException e) {
      return null;
    }
  }

  synchronized public SOFA.Connector.Reference getReference(String name) throws SOFA.SOFAnode.Run.Deployment.DeploymentException {
    if (!sroles.containsKey(name)) 
      throw new DeploymentException("No component with given name");
    return ((SOFA.Connector.RoleBase) sroles.get(name)).getSOFAReference();
  }


  /** Launching of depldock. Name of the depldock must be set by cmdline arguments.
    * If you want distributed SOFA node (i.e. depldock of one SOFA node run on several computers), you must
    * set "sofa.sofanode" property to the name of the SOFA node (it should be name of computer, where 
    * DeplDockRegistry runs. And also you should set properties  "sofa.rmihost", "sofa.rmiport", 
    * "sofa.orbport", "sofa.orbhost".
    * @param   argv   cmdline arguments<br>1st argument - name of the dock
    */
  public static void main(String[] argv) {
    if (argv.length != 1) {
      System.out.println("Set name of the dock");
      System.exit(1);
    }
    
    if (System.getSecurityManager() == null) {
      System.setSecurityManager(new java.rmi.RMISecurityManager());
    }

    String sofaNodeName = System.getProperty("sofa.nodename","");
    if (sofaNodeName.compareTo("")==0) {
      try {
        sofaNodeName = InetAddress.getLocalHost().getHostName();
      } catch (UnknownHostException e) { 
        System.out.println("Can't get host name. Set name to property \"sofa.nodename\"");
        System.exit(1);
      }
    }

    System.setProperty("sofa.nodename",sofaNodeName);

    SOFA.Util.CORBAAccess ca = new SOFA.Util.CORBAAccess();

    NameComponent nc = new NameComponent("DeplDockRegistry", "");
    NameComponent path[] = {nc};
    DeplDockRegistry ddr = null;
    try {
      ddr = DeplDockRegistryHelper.narrow(SOFA.Util.CORBAAccess.ncRef.resolve(path));
    } catch (Exception e) {
      System.out.println("Can't get DeplDock registry.");
      e.printStackTrace();
      System.exit(1);
    }

    DeplDockImpl ddi = new DeplDockImpl(argv[0], sofaNodeName, ddr);
    SOFA.Connector.RoleBase srole = null;
    try {
       srole = SOFA.Connector.Boot.DeplDockConnector.createSrv((DeplDock) ddi, argv[0]);
    } catch (SOFA.Connector.ConnectorException e) {
      System.out.println("Can't create connector to depldock");
      e.printStackTrace();
      System.exit(1);
    }
    SOFA.Connector.Reference ref = srole.getSOFAReference();
//    ref = removeLocalProfile(ref);
    byte[] strRef = sofaReferenceToBytes(ref);
//    System.out.println("###"+new String(strRef));
    ddr.register(strRef, argv[0]);
    System.out.println("DeplDock "+argv[0]+" is registered and running.");
    
    ca.run();
  }
}
