/* $Id: SOFACompClassLoader.java,v 1.1 2003/03/17 10:56:19 hnetynka Exp $ */
package SOFA.SOFAnode.Run;

import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/** SOFA component classloader. It loads classes of components.
  *
  * @author Petr Hnetynka
  */
public class SOFACompClassLoader  extends ClassLoader {
  
  /** URL with classes */
  URL url;

  /** Constructor.
    * @param _url      the URL from which to load classes
    */
  public SOFACompClassLoader(URL url) {
    super();
    this.url = url;
  }
  
  public Class findClass(String name) throws ClassNotFoundException {
    byte[] b;
    b = loadClassData(name);
    Class ret = defineClass(name, b, 0, b.length);
    return ret;
  }

  private byte[] loadClassData(String name) throws ClassNotFoundException {
    java.util.StringTokenizer tok = new java.util.StringTokenizer(name, ".");
    StringBuffer path = null;
    if (name.startsWith("SOFA.GeneratedConnectors")) {
      path = new StringBuffer(url.toString()+"/conn/");
    } else {
      path = new StringBuffer(url.toString()+"/classes/");
    }
    String nm = "";
    while (tok.hasMoreTokens()) {
      nm = tok.nextToken();
      path.append("/");
      path.append(nm);
    }
    path.append(".class");
    
    URL defURL;
    try {
      defURL = new URL(path.toString());
    } catch (MalformedURLException e) {
      throw new ClassNotFoundException("MalformedURLException: "+e.getMessage());
    }
    try {
      URLConnection con = defURL.openConnection();
      InputStream is = con.getInputStream();
      ByteArrayOutputStream bo = new ByteArrayOutputStream();
      int a = is.read();
      while (a != -1) {
        bo.write(a);
        a = is.read();
      }
      is.close();
      byte[] ar = bo.toByteArray();
      bo.close();
      return ar;
    } catch (IOException e) {
      throw new ClassNotFoundException("IOException: "+e.getMessage());
    }
  }



  private static SOFACompClassLoader classLoader;

  static {
    try {
      String baseURL = System.getProperty("sofa.tr.url","file:/sofa/tr/");
      classLoader = new SOFACompClassLoader(new URL(baseURL));
    } catch (java.net.MalformedURLException e) {
      throw new SOFARuntimeException(e.getMessage(), e);
    }
  }

  /** Returns sofa component class loader. There is one loader for all components in the dock.
    * @return class loader 
    */
  public static SOFACompClassLoader getCompClassLoader() {
    return classLoader;
  }

  /** Loades implementation class for component with given implementation version.
    * It is not for direct usage, it is used from component manager.
    * 
    * @see   SOFA.Component.ComponentManagerImpl#loadComponentClass(java.lang.String)  ComponentManagerImpl.loadComponentClass(String)
    * @param className        class name
    * @param compImplVersion  implementation version
    * @return                 loaded class
    * @exception ClassNotFoundException   class not found
    */
  public static Class loadComponentClass(String className, String compImplVersion) throws ClassNotFoundException {
    StringBuffer ret = new StringBuffer("_V_");
    for (int i=0; i<compImplVersion.length(); i++) {
      switch (compImplVersion.charAt(i)) {
      case '.': 
        ret.append('_');
        break;
      case '!':
        ret.append("_E_");
        break;
      default:
        ret.append(compImplVersion.charAt(i));
      }
    }
    return classLoader.loadClass(className+ret.toString());
  }
}
