/* $Id: TRLauncher.java,v 1.2 2004/03/22 19:31:49 hnetynka Exp $ */
package SOFA.SOFAnode.Run;

import SOFA.SOFAnode.Run.DeplDockRegistry.*;
import SOFA.SOFAnode.Run.Deployment.DeplDock;

import org.omg.CosNaming.*;

/** Launcher of SOFA applications - it takes depldescr from TR
  *
  * @author Petr Hnetynka
  */
public class TRLauncher {
 
  /** Launching of SOFA applications. Full application name must be set as a 1st cmdline argument,
    * 2nd argument is name of the home depldock. Set also all necessary properties
    * ("sofa.sofanode", "sofa.rmihost", "sofa.rmiport","sofa.orbport", "sofa.orbhost".
    * All depldocks necessary for application must run.
    */
  public static void main(String argv[]) {
    if (argv.length != 2) {
      System.out.println("Bad number of arguments.");
      System.out.println("Specify full application name and name of the home deployment dock");
      System.exit(1);
    }

    if (!argv[0].matches("^.+\\[.+\\]$")) {
      System.out.println("Bad format of application name.");
      System.exit(1);
    }

    if (System.getSecurityManager() == null) {
      System.setSecurityManager(new java.rmi.RMISecurityManager());
    }

    System.out.print("Obtaining and parsing deployment descriptor...");
    SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor depldesc = getDeplDescr(argv[0], "run.dd");
    System.out.println("OK");

    SOFA.Util.CORBAAccess ca = new SOFA.Util.CORBAAccess();

    NameComponent nc = new NameComponent("DeplDockRegistry", "");
    NameComponent path[] = {nc};
    DeplDockRegistry ddr = null;
    try {
      System.out.print("Obtaining depldock registry...");
      ddr = DeplDockRegistryHelper.narrow(SOFA.Util.CORBAAccess.ncRef.resolve(path));
      System.out.println("OK");
    } catch (Exception e) {
      System.out.println("Can't get DeplDock registry.");
      e.printStackTrace();
      System.exit(1);
    }
    
    byte[] serRef = null;
    try {
      System.out.print("Obtaining reference to depldock...");
      serRef = ddr.lookup(argv[1]);
      System.out.println("OK");
    } catch (NotFoundException e) {
      System.out.println("No deployment dock with name \""+argv[1]+"\".");
      System.exit(1);
    }
    
    SOFA.Connector.Reference ddRef = SOFA.SOFAnode.Run.Deployment.DeplDockImpl.bytesToSofaReference(serRef);
   
    DeplDock ddock = null;
    try {
      System.out.print("Creating connector to depldock...");
      ddock = (DeplDock) SOFA.Connector.Boot.DeplDockConnector.createClt(ddRef);
      System.out.println("OK");
    } catch (SOFA.Connector.ConnectorException e) {
      System.out.println("Can't create connector to deployment dock.");
      e.printStackTrace();
      System.exit(1);
    }
    
    SOFA.Connector.Reference apl = null;
    try {
      System.out.print("Instatiate component...");
      apl = ddock.instantiate(depldesc, "");
      System.out.println("OK");
    } catch (SOFA.SOFAnode.Run.Deployment.DeploymentException e) {
      System.out.println("DeploymentException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    }

    SOFA.Component.DCUP.DCUPComponentManager aplcm = null;
    try {
      System.out.print("Creating connector to aplication component manager...");
      aplcm = (SOFA.Component.DCUP.DCUPComponentManager) SOFA.Connector.Boot.DCUPComponentManagerConnector.createClt(apl);
      System.out.println("OK");
    } catch (SOFA.Connector.ConnectorException e) {
      System.out.println("Can't create connector to aplication component manager.");
      e.printStackTrace();
      System.exit(1);
    }
    
    try {
      System.out.print("Launching component...");
      aplcm.createComponent(new StorageImpl());
      System.out.println("OK");
    } catch (SOFA.Component.ComponentLifecycleException e) {
      System.out.println("ComponentLifecycleException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    }

    System.out.println(apl);
    System.out.println("Application is runnig :-)");
  }

  static SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor getDeplDescr(String applname, String ddname) {
    try {
      String trString = System.getProperty("sofa.tr.url","file:/sofa/tr/");
      String path = trString+"impl";
      path = path + fullNameToFileName(applname, ddname);
      java.net.URL url = new java.net.URL(path);
      java.net.URLConnection con = url.openConnection();
      java.io.InputStream is = con.getInputStream();
      SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor dd = new SOFA.SOFAnode.Run.Deployment.DeploymentDescriptorImpl(is);
      is.close();
      return dd;
    } catch (java.net.MalformedURLException e) {
      System.out.println("MalformedURLException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      System.out.println("ParserConfigurationExceptioni: "+e.getMessage() );
      System.exit(1);
    } catch (org.xml.sax.SAXException e) {
      System.out.println("SAXException: "+e.getMessage() );
      System.exit(1);
    }
    return null;
  }

  private static String fullNameToFileName(String clname, String ddname) {
    clname = clname.replaceAll("::", "/");
    clname = clname.replace('[', '/');
    clname = clname.replace(']', '/');
    clname = clname + ddname;
    if (clname.charAt(0) != '/')
      clname = "/" + clname;
    return clname;
  }
}
