/**
 * BundleOutputStream.java is a part of the SOFA project.
 * This file was created by pepan on 21.3.2003.
 */
package SOFA.SOFAnode.TR.Impl;

import SOFA.Util.Tools;

import java.io.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;

/**
 * A stream used for saving the contents of a bundle. It saves the bundle to a zipped
 * stream with a manifest file (JAR file).
 * @author Petr Panuska
 */
class BundleOutputStream extends JarOutputStream {

  /**
   * Creates this stream.
   * @param out the output stream lying under this one.
   * @throws IOException when some IO operation fails.
   */
  public BundleOutputStream (OutputStream out) throws IOException {
    super(out);
  }

  /**
   * Puts the next bundle entry (all files of a component) to this stream.
   * @param be the bundle entry being written to this stream.
   * @throws IOException when some IO operation fails.
   */
  public void putNextBundleEntry (BundleEntry be) throws IOException {
    Hashtable files = be.getFiles().getFiles();

    Enumeration paths = files.keys();
    while (paths.hasMoreElements()) {
      String path = (String) paths.nextElement();   // the path in the jar file
      File file = (File) files.get(path);            // the file in SOFAnode's file system
      long fileLength = file.length();

      path = path.replace('\\', '/');   // if there were some backslashes, in the JarEntry it must be a slash
      JarEntry jarEntry = new JarEntry(be.getName() + path); //location serves here as a prefix of path in a jar file!!!
      jarEntry.setSize(fileLength);
      jarEntry.setTime(file.lastModified());   // keep original time of creation
      putNextEntry(jarEntry);

      FileInputStream fileStream = new FileInputStream(file);

      Tools.copyStream(fileStream, this);

      fileStream.close();
      closeEntry();
    }
  }

}
