/**
 * Tools.java is a part of the SOFA project.
 * This file was created by pepan on 3.5.2003.
 */
package SOFA.Util;

import java.io.*;

/**
 * Contains a few useful commonly used tools.
 * @author  Petr Panuska
 */
public class Tools {
  /**
   * Deletes the given file. In case the file is a directory, it deletes all
   * sub-directories/files lying under this directory.
   * @param file the root directory or file which is going to be deleted.
   */
  public static void deleteSubTree (File file) {
    if (file.isDirectory()) {
      File[] subFiles = file.listFiles();
      for (int i = 0; i < subFiles.length; i++) {
        File subFile = subFiles[i];
        Tools.deleteSubTree(subFile); // recursively delete sub-files first
      }
    }
    file.delete();
  }

  /**
   * The length of a buffer used for reading data from a stream.
   */
  private static final int BUFFER_SIZE = 4096;

  /**
   * The buffer used for reading data from a stream.
   */
  private static final byte[] buffer = new byte[BUFFER_SIZE];

  /**
   * It copies all data from one stream to the second one using a static buffer.
   * Since the access to this buffer is synchronized, this method can be used
   * by many threads at one time with no problems.
   * @param src the input stream the data is read from.
   * @param dest the output stream the data is write to.
   * @throws IOException
   */
  public static void copyStream (InputStream src, OutputStream dest) throws IOException {
    int length;
    for (; ;) {
      synchronized (buffer) {
        length = src.read(buffer);
        if (length > 0)
          dest.write(buffer, 0, length);
        else
          break;
      }
    }
  }

}
